"""
:author: Niels Laukens
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xA5


@register(COMMAND_CODE, ["VMBDALI", "VMBDALI-20"])
class DimValueStatus(Message):
    """
    send by: VMBDALI
    received by:
    """

    def __init__(self, address: int = None) -> None:
        super().__init__()
        self.set_defaults(address)
        self.channel: int = 0
        self.dim_values: list[int] = []
        # dim_values contain dim value of channel, channel+1, ...

    def populate(self, priority, address: int, rtr: int, data: int) -> None:
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.set_attributes(priority, address, rtr)

        self.needs_data(data, 2)
        self.channel = data[0]
        self.dim_values = list(data[1:])

    def data_to_binary(self) -> bytes:
        return bytes(
            [
                COMMAND_CODE,
                self.channel,
            ]
        ) + bytes(self.dim_values)



"""
:author: Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xAC


@register(COMMAND_CODE)
class MemoTextMessage(Message):
    """
    send by:
    received by: VMBGPO, VMBGPOD
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.start = 0x00
        self.memo_text = ""
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.set_attributes(priority, address, rtr)
        self.start = data[1]
        self.name = "".join([chr(x) for x in data[2:]])

    def data_to_binary(self):
        """
        :return: bytes
        """
        while len(self.memo_text) < 5:
            self.memo_text += chr(0)
        return bytes([COMMAND_CODE, 0x00, self.start]) + bytes(self.memo_text, "utf-8")



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xF4


@register(COMMAND_CODE)
class UpdateLedStatusMessage(Message):
    """
    send by:
    received by: VMB6IN
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.led_on = []
        self.led_slow_blinking = []
        self.led_fast_blinking = []
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 3)
        self.set_attributes(priority, address, rtr)
        self.led_on = self.byte_to_channels(data[0])
        self.led_slow_blinking = self.byte_to_channels(data[1])
        self.led_fast_blinking = self.byte_to_channels(data[2])

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes(
            [
                COMMAND_CODE,
                self.channels_to_byte(self.led_on),
                self.channels_to_byte(self.led_slow_blinking),
                self.channels_to_byte(self.led_fast_blinking),
            ]
        )



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

import struct

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xFB
CHANNEL_NORMAL = 0x00
CHANNEL_INHIBITED = 0x01
CHANNEL_FORCED_ON = 0x02
CHANNEL_DISABLED = 0x03
RELAY_ON = 0x01
RELAY_OFF = 0x00
INTERVAL_TIMER_ON = 0x03
LED_OFF = 0
LED_ON = 1 << 7
LED_SLOW_BLINKING = 1 << 6
LED_FAST_BLINKING = 1 << 5
LED_VERY_FAST_BLINKING = 1 << 4


@register(COMMAND_CODE)
class RelayStatusMessage(Message):
    """
    send by: VMB4RYLD
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.channel = 0
        self.disable_inhibit_forced = 0
        self.status = 0
        self.led_status = 0
        self.delay_time = 0
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 7)
        self.set_attributes(priority, address, rtr)
        self.channel = self.byte_to_channel(data[0])
        self.disable_inhibit_forced = data[1]
        self.status = data[2]
        self.led_status = data[3]
        (self.delay_time,) = struct.unpack(">L", bytes([0]) + data[4:])

    def is_normal(self):
        """
        :return: bool
        """
        return self.disable_inhibit_forced == CHANNEL_NORMAL

    def is_inhibited(self):
        """
        :return: bool
        """
        return self.disable_inhibit_forced == CHANNEL_INHIBITED

    def is_forced_on(self):
        """
        :return: bool
        """
        return self.disable_inhibit_forced == CHANNEL_FORCED_ON

    def is_disabled(self):
        """
        :return: bool
        """
        return self.disable_inhibit_forced == CHANNEL_DISABLED

    def is_on(self):
        """
        :return: bool
        """
        return self.status == RELAY_ON

    def has_interval_timer_on(self):
        """
        :return: bool
        """
        return self.status == INTERVAL_TIMER_ON

    def data_to_binary(self):
        """
        :return: bytes
        """
        return (
            bytes(
                [
                    COMMAND_CODE,
                    self.channels_to_byte([self.channel]),
                    self.disable_inhibit_forced,
                    self.status,
                    self.led_status,
                ]
            )
            + struct.pack(">L", self.delay_time)[-3:]
        )


@register(COMMAND_CODE, ["VMB4RY"])
class RelayStatusMessage2(RelayStatusMessage):
    def is_on(self):
        """
        :return: bool
        """
        if (self.status >> (self.channel - 1)) & 1 != 0:
            return True
        else:
            return False



"""
:author: Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0x1C


@register(COMMAND_CODE, ["VMB1BLE", "VMB2BLE", "VMB1BLS", "VMB2BLE-10"])
class CoverPosMessage(Message):
    """
    sent by:
    received by: VMB2BLE
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.channel = 0
        self.position = 0
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_high_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 4)
        self.set_attributes(priority, address, rtr)
        self.channel = self.byte_to_channel(data[0])
        self.position = data[1]

    def set_defaults(self, address):
        if address is not None:
            self.set_address(address)
        self.set_high_priority()
        self.set_no_rtr()

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes(
            [
                COMMAND_CODE,
                self.channels_to_byte([self.channel]),
                self.position,
            ]
        )



"""
:author: Danny De Gaspari
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xB9


@register(COMMAND_CODE)
class TempSensorSettingsPart4(Message):
    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.set_attributes(priority, address, rtr)

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes([COMMAND_CODE])



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0x0A


@register(COMMAND_CODE)
class BusActiveMessage(Message):
    def set_defaults(self, address):
        if address is not None:
            self.set_address(address)
        self.set_high_priority()
        self.set_no_rtr()

    def populate(self, priority, address, rtr, data):
        self.needs_high_priority(priority)
        self.needs_no_rtr(rtr)
        self.set_attributes(priority, address, rtr)
        self.needs_no_data(data)

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes([COMMAND_CODE])



from __future__ import annotations

import asyncio
import json
import socket
from typing import Tuple

Address = Tuple[str, int]


class VelbusDiscoveryProtocol(asyncio.DatagramProtocol):
    def __init__(self, target: Address):
        self.target = target

    def connection_made(self, transport: asyncio.transports.DatagramTransport) -> None:
        self.transport = transport
        sock = transport.get_extra_info("socket")
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        string = "Velbus Navigation Request"
        self.transport.sendto(string.encode(), self.target)

    def datagram_received(self, data: bytes | str, addr: Address) -> None:
        # data received: b'{"message": "Velbus Navigation Guidance", "hostname": "Velbus", "model": "signum18", "id": "7b95834e", "velbus_port": 27015, "velbus_auth": false}' ('192.168.1.9', 32767)
        try:
            json_data = json.loads(data)
        except Exception:
            return
        if all(
            key in json_data
            for key in (
                "message",
                "hostname",
                "model",
                "id",
                "velbus_port",
                "velbus_auth",
            )
        ):
            res = {
                "address": addr[0],
                "hostname": json_data["hostname"],
                "model": json_data["model"],
                "id": json_data["id"],
                "port": json_data["velbus_port"],
                "auth": json_data["velbus_auth"],
            }
            print("data received:", res)



"""
Velbus packet handler
:Author maikel punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

import asyncio
import importlib.resources
import json
import logging
import os
import pathlib
import pprint
import sys
from typing import TYPE_CHECKING, Awaitable, Callable

from aiofile import async_open

from velbusaio.command_registry import commandRegistry
from velbusaio.const import (
    SCAN_MODULEINFO_TIMEOUT_INITIAL,
    SCAN_MODULEINFO_TIMEOUT_INTERVAL,
    SCAN_MODULETYPE_TIMEOUT,
)
from velbusaio.helpers import h2, keys_exists
from velbusaio.message import Message
from velbusaio.messages.module_subtype import ModuleSubTypeMessage
from velbusaio.messages.module_type import ModuleType2Message, ModuleTypeMessage
from velbusaio.raw_message import RawMessage

if TYPE_CHECKING:
    from velbusaio.controller import Velbus


class PacketHandler:
    """
    The packetHandler class
    """

    def __init__(
        self,
        velbus: Velbus,
        one_address: int | None = None,
    ) -> None:
        self._log = logging.getLogger("velbus-handler")
        self._log.setLevel(logging.DEBUG)
        self._velbus = velbus
        self._one_address = one_address
        self._typeResponseReceived = asyncio.Event()
        self._scanLock = asyncio.Lock()
        self._modulescan_address = 0
        self._scan_complete = False
        self._scan_delay_msec = 0

    async def read_protocol_data(self):
        if sys.version_info >= (3, 13):
            with importlib.resources.path(
                __name__, "module_spec/broadcast.json"
            ) as fspath:
                async with async_open(fspath) as protocol_file:
                    self.broadcast = json.loads(await protocol_file.read())
        else:
            async with async_open(
                str(
                    importlib.resources.files(__name__.split(".")[0]).joinpath(
                        "module_spec/broadcast.json"
                    )
                )
            ) as protocol_file:
                self.broadcast = json.loads(await protocol_file.read())

    def empty_cache(self) -> bool:
        if (
            len(
                [
                    name
                    for name in os.listdir(f"{self._velbus.get_cache_dir()}")
                    if os.path.isfile(f"{self._velbus.get_cache_dir()}/{name}")
                ]
            )
            == 0
        ):
            return True
        return False

    async def scan(self, reload_cache: bool = False) -> None:
        if reload_cache:
            self._modulescan_address = 0
            self._scan_complete = False
        # non-blocking check to see if the cache_dir is empty
        loop = asyncio.get_running_loop()
        if not reload_cache and await loop.run_in_executor(None, self.empty_cache):
            self._log.info("No cache yet, so forcing a bus scan")
            reload_cache = True
        self._log.info("Start module scan")
        while self._modulescan_address < 254:
            address = 0
            module = None
            async with self._scanLock:
                self._modulescan_address = self._modulescan_address + 1
                address = self._modulescan_address
                if self._velbus.addr_is_submodule(address):
                    self._log.info(
                        f"Skipping submodule address {address}, already handled"
                    )
                    continue
                self._log.info(f"Starting handling scan {address}")
                module = self._velbus.get_module(address)

            if self._one_address is not None and address != int(self._one_address):
                self._log.info(
                    f"Skipping address {address} as we requested to only scan one address {self._one_address}"
                )
                continue

            cfile = pathlib.Path(f"{self._velbus.get_cache_dir()}/{address}.json")
            # cleanup the old module cache if needed
            scanModule = reload_cache
            if scanModule and os.path.isfile(cfile):
                os.remove(cfile)
            elif os.path.isfile(cfile):
                scanModule = os.path.isfile(cfile)
            if scanModule:
                try:
                    self._log.info(f"Starting scan {address}")
                    self._typeResponseReceived.clear()
                    await self._velbus.sendTypeRequestMessage(address)
                    await asyncio.wait_for(
                        self._typeResponseReceived.wait(),
                        SCAN_MODULETYPE_TIMEOUT / 1000.0,
                    )
                    async with self._scanLock:
                        module = self._velbus.get_module(address)
                except asyncio.TimeoutError:
                    self._log.info(
                        f"Scan module {address} failed: not present or unavailable"
                    )
                if module is not None:
                    try:
                        self._log.debug(
                            f"Module {module._address} detected: start loading"
                        )
                        await asyncio.wait_for(
                            module.load(from_cache=True),
                            SCAN_MODULEINFO_TIMEOUT_INITIAL / 1000.0,
                        )
                        self._scan_delay_msec = module.get_initial_timeout()
                        while (
                            self._scan_delay_msec > 50 and not await module.is_loaded()
                        ):
                            # self._log.debug(
                            #    f"\t... waiting {self._scan_delay_msec} is_loaded={await module.is_loaded()}"
                            # )
                            self._scan_delay_msec = self._scan_delay_msec - 50
                            await asyncio.sleep(0.05)
                        self._log.info(
                            f"Scan module {address} completed, module loaded={await module.is_loaded()}"
                        )
                    except asyncio.TimeoutError:
                        self._log.error(
                            f"Module {address} did not respond to info requests after successful type request"
                        )
        self._scan_complete = True
        self._log.info("Module scan completed")

    async def handle(self, rawmsg: RawMessage) -> None:
        """
        Handle a received packet
        """
        if rawmsg.address < 1 or rawmsg.address > 254:
            return
        if rawmsg.command is None:
            return

        priority = rawmsg.priority
        address = rawmsg.address
        rtr = rawmsg.rtr
        command_value = rawmsg.command
        data = rawmsg.data_only

        # handle module type response message
        if command_value == 0xFF and not self._scan_complete:
            tmsg: ModuleTypeMessage = ModuleTypeMessage()
            tmsg.populate(priority, address, rtr, data)
            async with self._scanLock:
                await self._handle_module_type(tmsg)
                if address == self._modulescan_address:
                    self._typeResponseReceived.set()
                else:
                    self._log.debug(
                        f"Unexpected module type message module address {address}, Velbuslink scan?"
                    )
                    self._modulescan_address = address - 1

            self._typeResponseReceived.set()

        # handle module subtype response message
        elif command_value in (0xB0, 0xA7, 0xA6) and not self._scan_complete:
            msg: ModuleSubTypeMessage = ModuleSubTypeMessage()
            msg.populate(priority, address, rtr, data)
            if command_value == 0xB0:
                msg.sub_address_offset = 0
            elif command_value == 0xA7:
                msg.sub_address_offset = 4
            elif command_value == 0xA6:
                msg.sub_address_offset = 8
            async with self._scanLock:
                self._scan_delay_msec += SCAN_MODULEINFO_TIMEOUT_INTERVAL
                self._handle_module_subtype(msg)

        # ignore broadcast
        elif command_value in self.broadcast:
            self._log.debug(
                "Received broadcast message {} from {}, ignoring".format(
                    self.broadcast[str(command_value).upper()], address
                )
            )

        # handle other messages for modules that are already scanned
        else:
            module = None
            async with self._scanLock:
                module = self._velbus.get_module(address)
            if module is not None:
                module_type = module.get_type()
                if commandRegistry.has_command(int(command_value), module_type):
                    command = commandRegistry.get_command(command_value, module_type)
                    if not command:
                        return
                    msg = command()
                    msg.populate(priority, address, rtr, data)
                    # restart the info completion time when info message received
                    if command_value in (
                        0xF0,
                        0xF1,
                        0xF2,
                        0xFB,
                        0xFE,
                        0xCC,
                    ):  # names, memory data, memory block
                        self._scan_delay_msec += SCAN_MODULEINFO_TIMEOUT_INTERVAL
                        # self._log.debug(f"Restart timeout {msg}")
                    # send the message to the modules
                    await module.on_message(msg)
                else:
                    self._log.warning(f"NOT FOUND IN command_registry: {rawmsg}")

    async def _handle_module_type(
        self, msg: ModuleTypeMessage | ModuleType2Message
    ) -> None:
        """
        load the module data
        """
        if msg is not None:
            module = self._velbus.get_module(msg.address)
            if module is None:
                # data = keys_exists(self.pdata, "ModuleTypes", h2(msg.module_type))
                # if not data:
                #    self._log.warning(f"Module not recognized: {msg.module_type}")
                #    return
                await self._velbus.add_module(
                    msg.address,
                    msg.module_type,
                    memorymap=msg.memory_map_version,
                    build_year=msg.build_year,
                    build_week=msg.build_week,
                    serial=msg.serial,
                )
            else:
                self._log.debug(
                    f"***Module already exists scanAddr={self._modulescan_address} addr={msg.address} {msg}"
                )

        # else:
        #    self._log.debug("*** handle_module_type called without response message")

    def _handle_module_subtype(self, msg: ModuleSubTypeMessage) -> None:
        module = self._velbus.get_module(msg.address)
        if module is not None:
            addrList = {
                (msg.sub_address_offset + 1): msg.sub_address_1,
                (msg.sub_address_offset + 2): msg.sub_address_2,
                (msg.sub_address_offset + 3): msg.sub_address_3,
                (msg.sub_address_offset + 4): msg.sub_address_4,
            }
            self._velbus.add_submodules(module, addrList)


#    def _channel_convert(self, module: str, channel: str, ctype: str) -> None | int:
#        data = keys_exists(
#            self.pdata, "ModuleTypes", h2(module), "ChannelNumbers", ctype
#        )
#        if data and "Map" in data and h2(channel) in data["Map"]:
#            return data["Map"][h2(channel)]
#        if data and "Convert" in data:
#            return int(channel)
#        for offset in range(0, 8):
#            if channel & (1 << offset):
#                return offset + 1
#        return None



""":author: Frank van Breugel"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0x07


@register(
    COMMAND_CODE,
    ["VMB1DM", "VMBDME", "VMB4DC", "VMB1LED"],
)
class SetDimmerMessage(Message):
    """with this message the channel numbering is a bitnumber
    send by:
    received by: VMBDME, VMB4DC
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.dimmer_channels = []
        self.dimmer_state = 0
        self.dimmer_transitiontime = 0
        self.set_defaults(address)

    def set_defaults(self, address):
        if address is not None:
            self.set_address(address)
        self.set_high_priority()
        self.set_no_rtr()

    def populate(self, priority, address, rtr, data):
        """:return: None"""
        self.needs_high_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 4)
        self.set_attributes(priority, address, rtr)
        self.dimmer_channels = self.byte_to_channels(data[0])
        self.dimmer_state = data[1]
        self.dimmer_transitiontime = int.from_bytes(
            data[2:3], byteorder="big", signed=False
        )

    def data_to_binary(self):
        """:return: bytes"""
        return bytes(
            [
                COMMAND_CODE,
                self.channels_to_byte(self.dimmer_channels),
                self.dimmer_state,
            ]
        ) + self.dimmer_transitiontime.to_bytes(2, byteorder="big", signed=False)


@register(
    COMMAND_CODE,
    ["VMBDALI", "VMBDALI-20", "VMBDMI", "VMBDMI-R", "VMB8DC-20", "VMB4LEDPWM-20"],
)
class SetDimmerMessage2(SetDimmerMessage):
    """This with this message the channel numbering is an integer
    send by:
    received by: VMBDALI
    """

    def byte_to_channels(self, byte: int) -> list[int]:
        return [byte]

    def channels_to_byte(self, channels) -> int:
        if len(channels) != 1:
            raise ValueError("We should have exact one channel")
        return channels[0]



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

import struct

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xFF
MODULES_WITHOUT_SERIAL = {
    0x01: "VMB8PB",
    0x02: "VMB1RY",
    0x03: "VMB1BL",
    0x05: "VMB6IN",
    0x07: "VMB1DM",
    0x08: "VMB4RY",
    0x09: "VMB2BL",
    0x0C: "VMB1TS",
    0x0D: "VMB1TH",
    0x0E: "VMB1TC",
    0x0F: "VMB1LED",
    0x14: "VMBDME",
}


@register(
    COMMAND_CODE,
    [
        "VMB1BL",
        "VMB6IN",
        "VMB1DM",
        "VMB4RY",
        "VMB2BL",
        "VMB8IR",
        "VMB4PD",
        "VMB1TS",
        "VMB1TH",
        "VMB1TC",
        "VMB1LED",
        "VMB4RYLD",
        "VMB4RYNO",
        "VMB4DC",
        "VMBLCDWB",
        "VMBDME",
        "VMBDMI",
        "VMB8PBU",
        "VMB6PBN",
        "VMB2PBN",
        "VMB6PBB",
        "VMB4RF",
        "VMB1RYNO",
        "VMB1BLE",
        "VMB2BLE",
        "VMBGP1",
        "VMBGP2",
        "VMBGP4",
        "VMBGPO",
        "VMB7IN",
        "VMBGPOD",
        "VMB1RYNOS",
        "VMBPIRM",
        "VMBPIRC",
        "VMBPIRO",
        "VMBGP4PIR",
        "VMB1BLS",
        "VMBDMI-R",
        "VMBMETEO",
        "VMB4AN",
        "VMBVP01",
        "VMBEL1",
        "VMBEL2",
        "VMBEL4",
        "VMBELO",
        "VMBELPIR",
        "VMBSIG",
        "VMBGP1-2",
        "VMBGP2-2",
        "VMBGP4-2",
        "VMBGPOD-2",
        "VMBGP4PIR-2",
        "VMCM3",
        "VMBUSBIP",
        "VMB1RYS",
        "VMBKP",
        "VMBIN",
        "VMB4PB",
        "VMBDALI",
    ],
)
class ModuleTypeMessage(Message):
    """
    send by: VMB6IN, VMB4RYLD
    received by:
    """

    # pylint: disable-msg=R0902

    def __init__(self, address=None) -> None:
        Message.__init__(self)
        self.module_type = 0x00
        self.led_on = []
        self.led_slow_blinking = []
        self.led_fast_blinking = []
        self.serial = 0
        self.memory_map_version = 0
        self.build_year = 0
        self.build_week = 0
        self.set_defaults(address)

    def module_name(self) -> str:
        """
        :return: str
        """
        return "Unknown"

    def populate(self, priority, address, rtr, data) -> None:
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.set_attributes(priority, address, rtr)
        self.module_type = data[0]
        if data[0] not in MODULES_WITHOUT_SERIAL:
            (self.serial,) = struct.unpack(">L", bytes([0, 0, data[1], data[2]]))
            self.memory_map_version = data[3]
        self.build_year = data[-2]
        self.build_week = data[-1]


@register(
    COMMAND_CODE,
    [
        "VMB4RYLD-10",
        "VMB4RYNO-10",
        "VMB2BLE-10",
        "VMB6PB-20",
        "VMBEL1-20",
        "VMBEL2-20",
        "VMBEL4-20",
        "VMBELO-20",
        "VMBGP1-20",
        "VMBGP2-20",
        "VMBGP4-20",
        "VMBGPO-20",
        "VMBEL4PIR-20",
        "VMBGP4PIR-20",
        "VMB4LEDPWM-20",
        "VMB8IN-20",
        "VMB8DC-20",
    ],
)
class ModuleType2Message(Message):
    def __init__(self, address=None) -> None:
        Message.__init__(self)
        self.module_type = 0x00
        self.led_on = []
        self.led_slow_blinking = []
        self.led_fast_blinking = []
        self.serial = 0
        self.memory_map_version = 0
        self.build_year = 0
        self.build_week = 0
        self.term = 0
        self.set_defaults(address)

    def module_name(self) -> str:
        """
        :return: str
        """
        return "Unknown"

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.set_attributes(priority, address, rtr)
        self.module_type = data[0]
        if data[0] not in MODULES_WITHOUT_SERIAL:
            (self.serial,) = struct.unpack(">L", bytes([0, 0, data[1], data[2]]))
            self.memory_map_version = data[3]
        self.build_year = data[-3]
        self.build_week = data[-2]
        self.term = data[-1]



"""
:author: Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xA4


@register(COMMAND_CODE, ["VMB8IN-20"])
class CounterValueMessage(Message):

    def __init__(self, address=None):
        Message.__init__(self)
        self.channel = 0
        self.power = 0
        self.energy = 0

    def populate(self, priority, address, rtr, data):
        """
        -DB0    bit 0-4      = channel
        -DB0   bit 5-7      = Highest nibble (bits 19…16) of Power
        -DB1                 = bits 15…8 of Power
        -DB2                 = bits 7…0 of Power
        -DB3    bit 0-7      = energy counter
        -DB4    bit 0-7      = energy counter
        -DB5    bit 0-7      = energy counter
        :return: None
        """
        self.needs_no_rtr(rtr)
        self.needs_data(data, 7)
        self.set_attributes(priority, address, rtr)
        self.channel = (data[0] & 0x0F0) + 1
        self.power = (data[0] << 16) + (data[1] << 8) + data[2]
        self.energy = (data[3] << 16) + (data[4] << 8) + data[5]

    def get_channels(self):
        """
        :return: list
        """
        return self.channel



"""
:author: Danny De Gaspari
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xB3


@register(COMMAND_CODE)
class SelectProgramMessage(Message):
    def __init__(self, address=None, program=0):
        Message.__init__(self)
        self.select_program = program
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 1)
        self.set_attributes(priority, address, rtr)

        self.select_program = data[0] & 0x03

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes([COMMAND_CODE, self.select_program])



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

import struct

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xB0
COMMAND_CODE_2 = 0xA7
COMMAND_CODE_3 = 0xA6


@register(COMMAND_CODE)
@register(COMMAND_CODE_2)
@register(COMMAND_CODE_3)
class ModuleSubTypeMessage(Message):
    """
    send by: VMB6IN, VMB4RYLD
    received by:
    """

    # pylint: disable-msg=R0902

    def __init__(self, address=None, sub_address_offset: int = 0) -> None:
        Message.__init__(self)
        self.module_type = 0x00
        self.sub_address_1 = 0xFF
        self.sub_address_2 = 0xFF
        self.sub_address_3 = 0xFF
        self.sub_address_4 = 0xFF
        self.set_defaults(address)
        self.serial = 0
        self.sub_address_offset = sub_address_offset

    def module_name(self) -> str:
        """
        :return: str
        """
        return "Unknown"

    def populate(self, priority, address, rtr, data) -> None:
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        # self.needs_data(data, 6)
        self.set_attributes(priority, address, rtr)
        self.module_type = data[0]
        (self.serial,) = struct.unpack(">L", bytes([0, 0, data[1], data[2]]))
        self.sub_address_1 = data[3]
        self.sub_address_2 = data[4]
        self.sub_address_3 = data[5]
        self.sub_address_4 = data[6]



"""
Some common utils.
"""

from typing import Union

from velbusaio.const import MAXIMUM_MESSAGE_SIZE, MINIMUM_MESSAGE_SIZE


# Copyright (c) 2017 Thomas Delaet
# Copied from python-velbus (https://github.com/thomasdelaet/python-velbus)
def checksum(data: Union[bytes, bytearray]) -> int:
    if len(data) < MINIMUM_MESSAGE_SIZE - 2:
        raise ValueError("The message is shorter then expected")
    if len(data) > MAXIMUM_MESSAGE_SIZE - 2:
        raise ValueError("The message is longer then expected")
    __checksum = 0
    for data_byte in data:
        __checksum += data_byte
    __checksum = -(__checksum % 256) + 256
    return __checksum % 256


class VelbusException(Exception):
    """Velbus Exception."""

    def __init__(self, value):
        Exception.__init__(self)
        self.value = value

    def __str__(self) -> str:
        return repr(self.value)


class MessageParseException(Exception):
    pass


class BitSet:
    def __init__(self, value: int):
        self._value = value

    def __getitem__(self, idx: int) -> bool:
        if idx > 8 or idx <= 0:
            raise ValueError("The bitSet id is not within expected range 0 < id < 8")
        return bool((1 << idx) & self._value)

    def __setitem__(self, idx: int, value: bool) -> None:
        if idx > 8 or idx <= 0:
            raise ValueError("The bitSet id is not within expected range 0 < id < 8")
        mask = (0xFF ^ (1 << idx)) & self._value
        self._value = mask & (value << idx)

    def __len__(self) -> int:
        return 8  # a bitset represents one byte



"""
:author: Niels Laukens
"""

from __future__ import annotations

import dataclasses
import enum

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xE8


@register(COMMAND_CODE, ["VMBDALI", "VMBDALI-20"])
class DaliDeviceSettingMsg(Message):
    """
    send by: VMBDALI
    received by:
    """

    def __init__(self, address: int | None = None):
        super().__init__()
        self.set_defaults(address)
        self.channel: int = 0

    def populate(self, priority, address: int, rtr: int, data: bytes) -> None:
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.set_attributes(priority, address, rtr)

        self.needs_data(data, 2)
        self.channel = data[0]
        message_subtype = data[1]
        self.data: bytes | DaliDeviceSettingSubMessage = data[2:]
        try:
            decode_class = DaliDeviceSetting(message_subtype).decode_class
            if decode_class is not None:
                self.data = decode_class.from_data(self.data)
        except Exception:
            # Unknown subtype or error while decoding; leave as bytes
            pass

    def to_json_basic(self):
        me = {
            "name": str(self.__class__.__name__),
            "priority": self.priority,
            "rtr": self.rtr,
            "channel": self.channel,
        }
        if isinstance(self.data, DaliDeviceSettingSubMessage):
            me["data"] = self.data.to_json_basic()
        else:
            me["data"] = self.data.hex()
        return me


class DaliDeviceSettingSubMessage:
    """Abstract base class."""

    @classmethod
    def from_data(cls, data: bytes) -> DaliDeviceSettingSubMessage:
        raise NotImplementedError()

    def to_data(self) -> bytes:
        raise NotImplementedError()

    def to_json_basic(self):
        raise NotImplementedError()


class DeviceType(enum.Enum):
    FluorecentLamp = 0
    EmergencyLamp = 1
    DischargeLamp = 2
    LowVoltageLamp = 3
    Dimmer = 4
    ConversionToDc = 5
    LedModule = 6
    Relay = 7
    ColorControl = 8
    Sequencer = 9
    DevicePresent = 254
    NoDevicePresent = 255


@dataclasses.dataclass
class DeviceTypeMsg(DaliDeviceSettingSubMessage):
    device_type: DeviceType

    @classmethod
    def from_data(cls, data: bytes) -> DeviceTypeMsg:
        if len(data) != 1:
            raise ValueError("Expected 1 byte of data")
        return DeviceTypeMsg(device_type=DeviceType(data[0]))

    def to_data(self) -> bytes:
        return bytes([self.device_type.value])

    def to_json_basic(self):
        return {
            "submsg_type": self.__class__.__name__,
            "device_type": self.device_type.name,
        }


@dataclasses.dataclass
class MemberOfGroupMsg(DaliDeviceSettingSubMessage):
    member_of_group: list[bool]

    @classmethod
    def from_data(cls, data: bytes) -> DaliDeviceSettingSubMessage:
        if len(data) != 2:
            raise ValueError("Expected 2 bytes")
        i = int.from_bytes(data, "little")
        member_of_groups = []
        for group_num in range(0, 16):
            member_of_groups.append(bool(i & (1 << group_num)))
        return MemberOfGroupMsg(member_of_groups)

    def to_data(self) -> bytes:
        i = 0
        for group_num, member in enumerate(self.member_of_group):
            if member:
                i += 1 << group_num
        return i.to_bytes(2, "little")

    def to_json_basic(self):
        return {
            "member_of_groups": self.member_of_groups,
        }

    @property
    def member_of_groups(self) -> list[int]:
        groups = []
        for num, member in enumerate(self.member_of_group):
            if member:
                groups.append(num)
        return groups


class DaliDeviceSetting(enum.Enum):
    def __new__(
        cls, value: int, decode_class: type[DaliDeviceSettingSubMessage]
    ) -> DaliDeviceSetting:
        obj = object.__new__(cls)
        obj._value_ = value
        obj.decode_class = decode_class
        return obj

    Scene0Level = (0, None)
    Scene1Level = (1, None)
    Scene2Level = (2, None)
    Scene3Level = (3, None)
    Scene4Level = (4, None)
    Scene5Level = (5, None)
    Scene6Level = (6, None)
    Scene7Level = (7, None)
    Scene8Level = (8, None)
    Scene9Level = (9, None)
    Scene10Level = (10, None)
    Scene11Level = (11, None)
    Scene12Level = (12, None)
    Scene13Level = (13, None)
    Scene14Level = (14, None)
    Scene15Level = (15, None)
    PowerOnLevel = (16, None)
    SystemFailureLevel = (17, None)
    MinimumLevel = (18, None)
    MaximumLevel = (19, None)
    FadeTimeAndRate = (20, None)
    MemberOfGroup = (21, MemberOfGroupMsg)
    GroupMembers0 = (22, None)
    GroupMembers32 = (22, None)
    # currently undefined 24
    DeviceType = (25, DeviceTypeMsg)
    ActualLevel = (26, None)



"""
:author: Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

import json

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xEA
DSTATUS = {0: "run", 2: "manual", 4: "sleep", 6: "disable"}
DMODE = {0: "safe", 16: "night", 32: "day", 64: "comfort"}


@register(COMMAND_CODE)
class TempSensorStatusMessage(Message):
    """
    send by: VMBGPOD
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.local_control = 0  # 0=unlocked, 1 =locked
        self.status_mode = 0  # DSTATUS
        self.status_str = "run"
        self.auto_send = 0  # 0=disabled
        self.mode = 0  # DMODE
        self.mode_str = "safe"
        self.cool_mode = False
        self.heater = False
        self.boost = False
        self.pump = False
        self.cooler = False
        self.alarm1 = False
        self.alarm2 = False
        self.alarm3 = False
        self.alarm4 = False
        self.current_temp = None
        self.target_temp = None
        self.sleep_timer = None

    def getCurTemp(self):
        return self.current_temp

    def populate(self, priority, address, rtr, data):
        """
        -DB1    last bit        = local_control
        -DB1    bit 2+3         = status_mode
        -DB1    bit 4           = auto send
        -DB1    bit 5+6+7       = mode
        -DB1    bit 8           = cool/heat
        -DB2                    = program (not used)
        -DB3    last bit        = heater
        -DB3    bit 2           = boost
        -DB3    bit 3           = pump
        -DB3    bit 4           = cooler
        -DB4    bit 5           = alarm 1
        -DB4    bit 6           = alarm 2
        -DB4    bit 7           = alarm 3
        -DB4    bit 8           = alarm 4
        -DB5    current temp    = current temp
        -DB6    target temp     = target temp
        -DB7-8  sleep timer     = 0=off >0=x min
        :return: None
        """
        self.needs_no_rtr(rtr)
        self.needs_data(data, 7)
        self.set_attributes(priority, address, rtr)

        self.local_control = data[0] & 0x01
        self.status_mode = data[0] & 0x06
        self.status_str = DSTATUS[self.status_mode]
        self.auto_send = data[0] & 0x08
        self.mode = data[0] & 0x70
        self.mode_str = DMODE[self.mode]
        self.cool_mode = (data[0] & 0x80) == 0x80

        self.heater = (data[2] & 0x01) == 0x01
        self.boost = (data[2] & 0x02) == 0x02
        self.pump = (data[2] & 0x04) == 0x04
        self.cooler = (data[2] & 0x08) == 0x08
        self.alarm1 = (data[2] & 0x10) == 0x10
        self.alarm2 = (data[2] & 0x20) == 0x20
        self.alarm3 = (data[2] & 0x40) == 0x40
        self.alarm4 = (data[2] & 0x80) == 0x80

        self.current_temp = data[3] / 2
        self.target_temp = data[4] / 2

        self.sleep_timer = (data[5] << 8) + data[6]

    def to_json(self):
        """
        :return: str
        """
        json_dict = self.to_json_basic()
        json_dict["local_control"] = self.local_control
        json_dict["status_mode"] = DSTATUS[self.status_mode]
        json_dict["auto_send"] = self.auto_send
        json_dict["mode"] = DMODE[self.mode]
        json_dict["cool_mode"] = self.cool_mode
        json_dict["heater"] = self.heater
        json_dict["boost"] = self.boost
        json_dict["pump"] = self.pump
        json_dict["cooler"] = self.cooler
        json_dict["alarm1"] = self.alarm1
        json_dict["alarm2"] = self.alarm2
        json_dict["alarm3"] = self.alarm3
        json_dict["alarm4"] = self.alarm4
        json_dict["current_temp"] = self.current_temp
        json_dict["target_temp"] = self.target_temp
        json_dict["sleep_timer"] = self.sleep_timer
        return json.dumps(json_dict)



"""
author: Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

import asyncio
import math
import string
from typing import TYPE_CHECKING, Any, Awaitable, Callable

from velbusaio.command_registry import commandRegistry
from velbusaio.const import (
    DEVICE_CLASS_ILLUMINANCE,
    DEVICE_CLASS_TEMPERATURE,
    ENERGY_KILO_WATT_HOUR,
    TEMP_CELSIUS,
    VOLUME_CUBIC_METER_HOUR,
    VOLUME_LITERS_HOUR,
)
from velbusaio.message import Message
from velbusaio.messages.edge_set_color import CustomColorPriority, SetEdgeColorMessage
from velbusaio.messages.module_status import PROGRAM_SELECTION

if TYPE_CHECKING:
    from velbusaio.module import Module


class Channel:
    """
    A velbus channel
    This is the basic abstract class of a velbus channel
    """

    def __init__(
        self,
        module: Module,
        num: int,
        name: str,
        nameEditable: bool,
        subDevice: bool,
        writer: Callable[[Message], Awaitable[None]],
        address: int,
    ):
        self._num = num
        self._module = module
        self._name = name
        self._subDevice = subDevice
        if not nameEditable:
            self._is_loaded = True
        else:
            self._is_loaded = False
        self._writer = writer
        self._address = address
        self._on_status_update = []
        self._name_parts = {}

    def get_module_type(self) -> int:
        return self._module.get_type()

    def get_module_type_name(self) -> str:
        return self._module.get_type_name()

    def get_module_serial(self) -> str:
        return self._module.get_serial()

    def get_module_address(self, chan_type: str = "") -> int:
        """Return (sub)module address for channel"""
        if chan_type == "Button" and self._num > 24:
            return self._module.get_addresses()[3]
        elif chan_type == "Button" and self._num > 16:
            return self._module.get_addresses()[2]
        elif chan_type == "Button" and self._num > 8:
            return self._module.get_addresses()[1]
        else:
            return self._address

    def get_module_sw_version(self) -> str:
        return self._module.get_sw_version()

    def get_channel_number(self) -> int:
        return self._num

    def get_full_name(self) -> str:
        if self._subDevice:
            return f"{self._module.get_name()} ({self._module.get_type_name()}) - {self._name}"
        return f"{self._module.get_name()} ({self._module.get_type_name()})"

    def is_loaded(self) -> bool:
        """
        Is this channel loaded
        """
        return self._is_loaded

    def is_counter_channel(self) -> bool:
        return False

    def is_temperature(self) -> bool:
        return False

    def is_sub_device(self) -> bool:
        return self._subDevice

    def get_name(self) -> str:
        """
        :return: the channel name
        """
        return self._name

    def set_name_char(self, pos: int, char: int) -> None:
        self._is_loaded = True
        self._name_parts = {}
        # make sure the string is long enough
        while len(self._name) < int(pos):
            self._name += " "
        # store the char on correct pos
        self._name = self._name[: int(pos)] + chr(char) + self._name[int(pos) + 1 :]

    def set_name_part(self, part: int, name: str) -> None:
        """
        Set a part of the channel name
        """
        # if int(part) not in self._name_parts:
        #    return
        self._name_parts[int(part)] = name
        if len(self._name_parts) == 3:
            self._generate_name()

    def _generate_name(self) -> None:
        """
        Generate the channel name if all 3 parts are received
        """
        name = self._name_parts[1] + self._name_parts[2] + self._name_parts[3]
        self._name = "".join(filter(lambda x: x in string.printable, name))
        self._is_loaded = True
        self._name_parts = {}

    def __getstate__(self):
        d = self.__dict__
        return {
            k: d[k]
            for k in d
            if k != "_writer" and k != "_on_status_update" and k != "_name_parts"
        }

    def to_cache(self) -> dict:
        dst = {
            "name": self._name,
            "type": type(self).__name__,
            "subdevice": self._subDevice,
        }
        if hasattr(self, "_Unit"):
            dst["Unit"] = self._Unit
        return dst

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._on_status_update = []
        self._name_parts = {}

    def __repr__(self) -> str:
        items = []
        for k, v in self.__dict__.items():
            if k not in ["_module", "_writer", "_name_parts"]:
                items.append(f"{k} = {v!r}")
        return "{}[{}]".format(type(self), ", ".join(items))

    def __str__(self) -> str:
        return self.__repr__()

    def get_channel_info(self) -> dict[str, Any]:
        data = {}
        for key, value in self.__dict__.items():
            data["type"] = self.__class__.__name__
            if key not in ["_module", "_writer", "_name_parts", "_on_status_update"]:
                data[key.replace("_", "", 1)] = value
        return data

    async def update(self, data: dict) -> None:
        """
        Set the attributes of this channel
        """
        for key, new_val in data.items():
            cur_val = getattr(self, f"_{key}", None)
            if cur_val is None or cur_val != new_val:
                setattr(self, f"_{key}", new_val)
                for m in self._on_status_update:
                    await m()

    def get_categories(self) -> list[str]:
        """
        Get the categories (mainly for home-assistant)
        """
        # COMPONENT_TYPES = ["switch", "sensor", "binary_sensor", "cover", "climate", "light"]
        return []

    def on_status_update(self, meth: Callable[[], Awaitable[None]]) -> None:
        self._on_status_update.append(meth)

    def remove_on_status_update(self, meth: Callable[[], Awaitable[None]]) -> None:
        self._on_status_update.remove(meth)

    def get_counter_state(self) -> int:
        raise NotImplementedError()

    def get_counter_unit(self) -> str:
        raise NotImplementedError()

    def get_max(self) -> int:
        raise NotImplementedError()

    def get_min(self) -> int:
        raise NotImplementedError()

    def is_water(self) -> bool:
        return False

    async def press(self) -> None:
        raise NotImplementedError()


class Blind(Channel):
    """
    A blind channel
    """

    _state = None
    # State reports the direction of *movement*: moving up, moving down or stopped
    _position = None
    # Position reporting is not supported by VMBxBL modules (only in BLE/BLS)

    def get_categories(self) -> list[str]:
        return ["cover"]

    def get_position(self) -> int | None:
        return self._position

    def get_state(self) -> str:
        return self._state

    def is_opening(self) -> bool:
        return self._state == 0x01

    def is_closing(self) -> bool:
        return self._state == 0x02

    def is_stopped(self) -> bool:
        return self._state == 0x00

    def is_closed(self) -> bool | None:
        """Report if the blind is fully closed."""
        if self._position is None:
            return None
        # else:
        return self._position == 100

    def is_open(self) -> bool | None:
        """Report if the blind is fully open."""
        if self._position is None:
            return None
        return self._position == 0

    def support_position(self) -> bool:
        # position will be populated after the first BlindStatusNgMessage (during module load)
        # For VMBxBL modules, position will remain None and not be overwritten
        return self._position is not None

    async def open(self) -> None:
        cls = commandRegistry.get_command(0x05, self._module.get_type())
        msg = cls(self._address)
        msg.channel = self._num
        await self._writer(msg)

    async def close(self) -> None:
        cls = commandRegistry.get_command(0x06, self._module.get_type())
        msg = cls(self._address)
        msg.channel = self._num
        await self._writer(msg)

    async def stop(self) -> None:
        cls = commandRegistry.get_command(0x04, self._module.get_type())
        msg = cls(self._address)
        msg.channel = self._num
        await self._writer(msg)

    async def set_position(self, position: int) -> None:
        # may not be supported by the module
        if position == 100:
            # at least VMB1BLS ignores command 0x1C with position 0x64
            await self.close()
            return
        cls = commandRegistry.get_command(0x1C, self._module.get_type())
        msg = cls(self._address)
        msg.channel = self._num
        msg.position = position
        await self._writer(msg)


class Button(Channel):
    """
    A Button channel
    """

    _enabled = True
    _closed = False
    _led_state = None
    _long = False

    def get_categories(self) -> list[str]:
        if self._enabled:
            return ["binary_sensor", "led", "button"]
        return []

    def is_closed(self) -> bool:
        """
        Return if this button is on
        """
        return self._closed

    def is_on(self) -> bool:
        """
        Return if this relay is on
        """
        if self._led_state == "on":
            return True
        return False

    async def set_led_state(self, state: str) -> None:
        """
        Set led
        """
        if state == "on":
            code = 0xF6
        elif state == "slow":
            code = 0xF7
        elif state == "fast":
            code = 0xF8
        elif state == "off":
            code = 0xF5
        else:
            return

        _mod_add = self.get_module_address("Button")
        _chn_num = self._num - self._module.calc_channel_offset(_mod_add)
        cls = commandRegistry.get_command(code, self._module.get_type())
        msg = cls(_mod_add)
        msg.leds = [_chn_num]
        await self._writer(msg)
        await self.update({"led_state": state})

    async def press(self) -> None:
        """
        Press the button
        """
        _mod_add = self.get_module_address("Button")
        _chn_num = self._num - self._module.calc_channel_offset(_mod_add)
        # send the just pressed
        cls = commandRegistry.get_command(0x00, self._module.get_type())
        msg = cls(_mod_add)
        msg.closed = [_chn_num]
        await self._writer(msg)
        # wait
        await asyncio.sleep(0.3)
        # send the just released
        msg = cls(_mod_add)
        msg.opened = [_chn_num]
        await self._writer(msg)


class ButtonCounter(Button):
    """
    A ButtonCounter channel
    This channel can act as a button and as a counter
    => standard     this is the calculated value
    => is_counter   this is the numeric value
    """

    _Unit = None
    _pulses = None
    _counter = None
    _delay = None
    _power = None
    _energy = None

    def get_categories(self) -> list[str]:
        if self._counter:
            return ["sensor"]
        return ["binary_sensor", "button"]

    def is_counter_channel(self) -> bool:
        if self._counter:
            return True
        return False

    def get_state(self) -> int:
        if self._energy:
            return self._energy
        # if we don't know the delay
        # or we don't know the unit
        # or the delay is the max value
        #   we always return 0
        val = 0
        if not self._delay or not self._Unit or self._delay == 0xFFFF:
            return round(0, 2)
        if self._Unit == VOLUME_LITERS_HOUR:
            val = (1000 * 3600) / (self._delay * self._pulses)
        elif self._Unit == VOLUME_CUBIC_METER_HOUR:
            val = (1000 * 3600) / (self._delay * self._pulses)
        elif self._Unit == ENERGY_KILO_WATT_HOUR:
            val = (1000 * 1000 * 3600) / (self._delay * self._pulses)
        else:
            val = 0
        return round(val, 2)

    def get_unit(self) -> str | None:
        if self._Unit == VOLUME_LITERS_HOUR:
            return "L"
        if self._Unit == VOLUME_CUBIC_METER_HOUR:
            return "m3"
        if self._Unit == ENERGY_KILO_WATT_HOUR:
            return "W"
        return None

    def get_counter_state(self) -> int:
        if self._power:
            return self._power
        return round((self._counter / self._pulses), 2)

    def get_counter_unit(self) -> str:
        return self._Unit

    def is_water(self) -> bool:
        if self._counter and self._Unit == VOLUME_LITERS_HOUR:
            return True
        return False


class Sensor(Button):
    """
    A Sensor channel
    This is a bit weird, but this happens because of code sharing with openhab
    A sensor in this case is actually a Button
    """

    def get_categories(self) -> list[str]:
        if self._enabled:
            return ["binary_sensor", "led"]
        return []


class ThermostatChannel(Button):
    """
    A Thermostat channel
    These are the booster/heater/alarms
    """


class Dimmer(Channel):
    """
    A Dimmer channel
    """

    _state: int = 0

    def __init__(
        self,
        module: Module,
        num: int,
        name: str,
        nameEditable: bool,
        subDevice: bool,
        writer: Callable[[Message], Awaitable[None]],
        address: int,
        slider_scale: int = 100,
    ):
        super().__init__(module, num, name, nameEditable, subDevice, writer, address)

        self.slider_scale = slider_scale
        # VMB4DC has dim values 0(off), 1-99(dimmed), 100(full on)
        # VMBDALI has dim values 0(off), 1-253(dimmed), 254(full on), 255(previous value)

    def get_categories(self) -> list[str]:
        return ["light"]

    def is_on(self) -> bool:
        """
        Check if a dimmer is turned on
        """
        if self._state == 0:
            return False
        return True

    def get_dimmer_state(self) -> int:
        """
        Return the dimmer state
        """
        return int(self._state * 100 / self.slider_scale)

    async def set_dimmer_state(self, slider: int, transitiontime: int = 0) -> None:
        """
        Set dimmer to slider
        """
        cls = commandRegistry.get_command(0x07, self._module.get_type())
        msg = cls(self._address)
        msg.dimmer_state = int(slider * self.slider_scale / 100)
        msg.dimmer_transitiontime = int(transitiontime)
        msg.dimmer_channels = [self._num]
        await self._writer(msg)

    async def restore_dimmer_state(self, transitiontime: int = 0) -> None:
        """
        restore dimmer to last known state
        """
        cls = commandRegistry.get_command(0x11, self._module.get_type())
        msg = cls(self._address)
        msg.dimmer_transitiontime = int(transitiontime)
        msg.dimmer_channels = [self._num]
        await self._writer(msg)


class Temperature(Channel):
    """
    A Temperature sensor channel
    """

    _cur = 0
    _cur_precision = None
    _max = None
    _min = None
    _target = 0
    _cmode = None
    _cool_mode = None
    _cstatus = None
    _thermostat = False
    _sleep_timer = 0

    def get_categories(self) -> list[str]:
        if self._thermostat:
            return ["sensor", "climate"]
        return ["sensor"]

    def get_class(self) -> str:
        return DEVICE_CLASS_TEMPERATURE

    def get_unit(self) -> str:
        return TEMP_CELSIUS

    def get_state(self) -> int:
        return round(self._cur, 2)

    def is_temperature(self) -> bool:
        return True

    def get_max(self) -> int | None:
        if self._max is None:
            return None
        return round(self._max, 2)

    def get_min(self) -> int | None:
        if self._min is None:
            return None
        return round(self._min, 2)

    def get_climate_target(self) -> int:
        return round(self._target, 2)

    def get_climate_preset(self) -> str:
        return self._cmode

    def get_climate_mode(self) -> str:
        return self._cstatus

    def get_cool_mode(self) -> str:
        return self._cool_mode

    async def set_temp(self, temp: float) -> None:
        cls = commandRegistry.get_command(0xE4, self._module.get_type())
        msg = cls(self._address)
        msg.temp = temp * 2  # TODO: int()
        await self._writer(msg)

    async def _switch_mode(self) -> None:
        if self._cmode == "safe":
            code = 0xDE
        elif self._cmode == "comfort":
            code = 0xDB
        elif self._cmode == "day":
            code = 0xDC
        else:  # "night"
            code = 0xDD

        if self._cstatus == "run":
            sleep = 0x0
        elif self._cstatus == "manual":
            sleep = 0xFFFF
        elif self._cstatus == "sleep":
            sleep = self._sleep_timer
        else:
            sleep = 0x0
        cls = commandRegistry.get_command(code, self._module.get_type())
        msg = cls(self._address, sleep)
        await self._writer(msg)

    async def set_preset(self, preset: str) -> None:
        self._cmode = preset
        await self._switch_mode()

    async def set_climate_mode(self, mode: str) -> None:
        self._cstatus = mode
        await self._switch_mode()

    async def set_mode(self, mode: str) -> None:
        # TODO: change function name, proposal = set_heat_cool_mode
        if mode == "heat":
            code = 0xE0
        elif mode == "cool":
            code = 0xDF
        # TODO: else case
        cls = commandRegistry.get_command(code, self._module.get_type())
        msg = cls(self._address)
        await self._writer(msg)

    async def maybe_update_temperature(self, new_temp: float, precision: float) -> None:
        # Based on experiments, Velbus modules seem to truncate (i.e. round down)
        current_temp_rounded_to_precision = (
            math.floor(self._cur / precision) * precision
        )

        if current_temp_rounded_to_precision == new_temp:
            # The newly received temperature is still in line with our current value,
            # but with reduced precision.
            # Don't update (would lose high precision)
            return

        elif (
            current_temp_rounded_to_precision - precision
            <= new_temp
            < current_temp_rounded_to_precision
            and self._cur_precision < precision
        ):
            # The newly received temperature is 1 LSb below the current value
            # and the current value was set by a better precision message
            # Modify the received temperature by "adding precision", while still keeping the same low precision value
            # e.g. (decimal digits represent precision)
            # | Actual  | Msg     | Stored  |
            # | 21.0000 | 21.0000 | 21.0000 |
            # | 20.9375 | 20.5    | 20.9375 |
            new_temp = current_temp_rounded_to_precision - self._cur_precision

        await self.update(
            {
                "cur": new_temp,
                "cur_precision": precision,
            }
        )


class SensorNumber(Channel):
    """
    A Numeric Sensor channel
    """

    _cur = 0
    _unit = None

    def get_categories(self) -> list[str]:
        return ["sensor"]

    def get_class(self) -> None:
        return None

    def get_unit(self) -> None:
        return self._unit

    def get_state(self) -> float:
        return round(self._cur, 2)


class LightSensor(Channel):
    """
    A light sensor channel
    """

    _cur = 0

    def get_categories(self) -> list[str]:
        return ["sensor"]

    def get_class(self) -> str:
        return DEVICE_CLASS_ILLUMINANCE

    def get_unit(self) -> None:
        return None

    def get_state(self) -> float:
        return round(self._cur, 2)


class Relay(Channel):
    """
    A Relay channel
    """

    _on = None
    _enabled = True
    _inhibit = False
    _forced_on = False
    _disabled = False

    def get_categories(self) -> list[str]:
        if self._enabled:
            return ["switch"]
        return []

    def is_on(self) -> bool:
        """
        Return if this relay is on
        """
        return self._on

    def is_inhibit(self) -> bool:
        return self._inhibit

    def is_forced_on(self) -> bool:
        return self._forced_on

    def is_disabled(self) -> bool:
        return self._disabled

    async def turn_on(self) -> None:
        """
        Send the turn on message
        """
        cls = commandRegistry.get_command(0x02, self._module.get_type())
        msg = cls(self._address)
        msg.relay_channels = [self._num]
        await self._writer(msg)

    async def turn_off(self) -> None:
        """
        Send the turn off message
        """
        cls = commandRegistry.get_command(0x01, self._module.get_type())
        msg = cls(self._address)
        msg.relay_channels = [self._num]
        await self._writer(msg)


class EdgeLit(Channel):
    """
    An EdgeLit channel
    """

    async def reset_color(self, left=True, top=True, right=True, bottom=True):
        msg = SetEdgeColorMessage(self._address)
        msg.apply_background_color = True
        msg.color_idx = 0
        msg.apply_to_left_edge = left
        msg.apply_to_top_edge = top
        msg.apply_to_right_edge = right
        msg.apply_to_bottom_edge = bottom
        msg.apply_to_all_pages = True
        await self._writer(msg)

    async def set_color(
        self,
        color_idx: int,
        left=True,
        top=True,
        right=True,
        bottom=True,
        blinking=False,
        priority=CustomColorPriority.LOW_PRIORITY,
    ) -> None:
        """
        Send the turn off message
        """

        msg = SetEdgeColorMessage(self._address)
        msg.apply_background_color = True
        msg.background_blinking = blinking
        msg.color_idx = color_idx
        msg.apply_to_left_edge = left
        msg.apply_to_top_edge = top
        msg.apply_to_right_edge = right
        msg.apply_to_bottom_edge = bottom
        msg.apply_to_all_pages = True
        msg.custom_color_priority = priority
        await self._writer(msg)


class Memo(Channel):
    """
    A Memo text
    """

    async def set(self, txt: str) -> None:
        cls = commandRegistry.get_command(0xAC, self._module.get_type())
        msg = cls(self._address)
        msgcntr = 0
        for char in txt:
            msg.memo_text += char
            if len(msg.memo_text) >= 5:
                msgcntr += 5
                await self._writer(msg)
                msg = cls(self._address)
                msg.start = msgcntr
        await self._writer(msg)


class SelectedProgram(Channel):
    """
    A selected program channel
    """

    _selected_program_str = None

    def get_categories(self) -> list[str]:
        return ["select"]

    def get_class(self) -> None:
        return None

    def get_options(self) -> list:
        return list(PROGRAM_SELECTION.values())

    def get_selected_program(self) -> str:
        return self._selected_program_str

    async def set_selected_program(self, program_str: str) -> None:
        self._selected_program_str = program_str
        command_code = 0xB3
        cls = commandRegistry.get_command(command_code, self._module.get_type())
        index = list(PROGRAM_SELECTION.values()).index(program_str)
        program = list(PROGRAM_SELECTION.keys())[index]
        msg = cls(self._address, program)
        await self._writer(msg)



"""
:author: Tom Dupré <gitd8400@gmail.com>
"""

from __future__ import annotations

import json

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xEC
DSTATUS = {0: "off", 1: "up", 2: "down"}


@register(COMMAND_CODE, ["VMB1BLE", "VMB2BLE", "VMB1BLS", "VMB2BLE-10"])
class BlindStatusNgMessage(Message):
    """
    sent by: VMB2BLE
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.channel = 0
        self.timeout = 0
        self.status = 0
        self.position = None
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 7)
        self.set_attributes(priority, address, rtr)
        self.channel = self.byte_to_channel(data[0])
        self.timeout = data[1]  # Omzetter seconden ????
        self.status = data[2]
        self.position = data[4]

    def to_json(self):
        """
        :return: str
        """
        json_dict = self.to_json_basic()
        json_dict["channel"] = self.channel
        json_dict["timeout"] = self.timeout
        json_dict["status"] = DSTATUS[self.status]
        return json.dumps(json_dict)

    def is_moving_up(self) -> bool:
        return self.status == 0x01

    def is_moving_down(self) -> bool:
        return self.status == 0x02

    def is_stopped(self) -> bool:
        return self.status == 0x00

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes(
            [
                COMMAND_CODE,
                self.channels_to_byte([self.channel]),
                self.timeout,
                self.status,
                self.led_status,
                self.blind_position,
                self.locked_inhibit_forced,
                self.alarm_auto_mode_selection,
            ]
        )


@register(COMMAND_CODE, ["VMB1BL", "VMB2BL"])
class BlindStatusMessage(Message):
    """
    sent by: VMB2BLE
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.channel = 0
        self.timeout = 0
        self.status = 0
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 7)
        self.set_attributes(priority, address, rtr)
        # 00000011 = channel 1
        # 00001100 = channel 2
        # so shift 1 bit to the right + and with 03
        tmp = (data[0] >> 1) & 0x03
        self.channel = self.byte_to_channel(tmp)
        self.timeout = data[1]  # Omzetter seconden ????
        # 2 bits per channel used
        self.status = (data[2] >> ((self.channel - 1) * 2)) & 0x03

    def to_json(self):
        """
        :return: str
        """
        json_dict = self.to_json_basic()
        json_dict["channel"] = self.channel
        json_dict["timeout"] = self.timeout
        json_dict["status"] = DSTATUS[self.status]
        return json.dumps(json_dict)

    def is_moving_up(self) -> bool:
        return self.status == 0x01

    def is_moving_down(self) -> bool:
        return self.status == 0x02

    def is_stopped(self) -> bool:
        return self.status == 0x00



"""
:author: Frank van Breugel
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0x0F


@register(
    COMMAND_CODE,
    [
        "VMB1DM",
        "VMBDME",
        "VMB4DC",
        "VMBDMI",
        "VMBDMI-R",
        "VMB1LED",
        "VMB8DC-20",
        "VMB4LEDPWM-20",
    ],
)
class SliderStatusMessage(Message):
    """
    sent by: VMBDME
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.channel = 0
        self.slider_state = 0
        self.slider_long_pressed = 0
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_high_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 3)
        self.set_attributes(priority, address, rtr)
        self.channel = self.byte_to_channel(data[0])
        self.slider_state = int.from_bytes([data[1]], byteorder="big")
        self.slider_long_pressed = data[2]

    def cur_slider_state(self):
        """
        :return: int
        """
        return self.slider_state

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes(
            [
                COMMAND_CODE,
                self.channels_to_byte([self.channel]),
                self.slider_state,
                self.slider_long_pressed,
            ]
        )



"""
:author: Tom Dupré <gitd8400@gmail.com>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0x04


@register(COMMAND_CODE, ["VMB1BLE", "VMB2BLE", "VMB1BLS", "VMB2BLE-10"])
class CoverOffMessage(Message):
    """
    sent by:
    received by: VMB2BLE
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.channel = 0
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_high_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 1)
        self.set_attributes(priority, address, rtr)
        self.channel = self.byte_to_channel(data[0])

    def set_defaults(self, address):
        if address is not None:
            self.set_address(address)
        self.set_high_priority()
        self.set_no_rtr()

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes([COMMAND_CODE, self.channels_to_byte([self.channel])])


@register(COMMAND_CODE, ["VMB1BL", "VMB2BL"])
class CoverOffMessage2(Message):
    """
    sent by:
    received by: VMB1BL VMB2BL
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.channel = 0
        self.delay_time = 0
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_high_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 1)
        self.set_attributes(priority, address, rtr)
        # 00000011 = channel 1
        # 00001100 = channel 2
        # so shift 1 bit to the right + and with 03
        tmp = (data[0] >> 1) & 0x03
        self.channel = self.byte_to_channel(tmp)

    def set_defaults(self, address):
        if address is not None:
            self.set_address(address)
        self.set_high_priority()
        self.set_no_rtr()

    def data_to_binary(self):
        """
        :return: bytes
        """
        if self.channel == 0x01:
            tmp = 0x03
        else:
            tmp = 0x0C

        return bytes([COMMAND_CODE, tmp])



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0x0E


@register(COMMAND_CODE)
class InterfaceStatusRequestMessage(Message):
    """
    send by: VMB1USB
    received by:
    """

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_no_data(data)
        self.set_attributes(priority, address, rtr)

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes([COMMAND_CODE])



"""
:author: Maikel Punie
"""

from __future__ import annotations

import struct

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xA2
BALANCED = 0x01
BOOST = 0x02
BACKUP = 0x03


@register(COMMAND_CODE, ["VMBPSUMNGR-20"])
class PsuLoadMessage(Message):

    def __init__(self, address=None):
        Message.__init__(self)
        self.mode = 0
        self.load_1 = 0
        self.load_2 = 0
        self.out = 0

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 3)
        self.set_attributes(priority, address, rtr)
        self.mode = data[0]
        self.load_1 = data[1]
        self.load_2 = data[2]
        self.out = data[3]

    def data_to_binary(self):
        """
        :return: bytes
        """
        return (
            bytes(
                [
                    COMMAND_CODE,
                    self.mode,
                    self.load_1,
                    self.load_2,
                    self.out,
                ]
            )
            + struct.pack(">L", self.delay_time)[-3:]
        )



"""
:author: Frank van Breugel
"""

from __future__ import annotations

import struct

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xEE
MODE_START_STOP = 0x00
MODE_STAIRCASE = 0x01
MODE_DIMMER = 0x02
MODE_MEMORY = 0x03
MODE_MULTI = 0x04
MODE_SLOW_ON = 0x05
MODE_SLOW_OFF = 0x06
MODE_SLOW = 0x06
LED_OFF = 0
LED_ON = 1 << 7
LED_SLOW_BLINKING = 1 << 6
LED_FAST_BLINKING = 1 << 5
LED_VERY_FAST_BLINKING = 1 << 4


@register(COMMAND_CODE, ["VMB1DM", "VMBDME", "VMB1LED"])
class DimmerStatusMessage(Message):
    """
    sent by: VMBDME
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.channel = 1
        self.disable_inhibit_forced = 0
        self.dimmer_mode = 0
        self.dimmer_state = 0
        self.led_status = 0
        self.delay_time = 0
        self.set_defaults(address)
        self.dimmer_config = 0

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 7)
        self.set_attributes(priority, address, rtr)
        self.dimmer_mode = data[0]
        self.channel = 1
        self.dimmer_state = int.from_bytes([data[1]], byteorder="big", signed=False)
        self.led_status = data[2]
        (self.delay_time,) = struct.unpack(">L", bytes([0]) + data[4:])
        self.dimmer_config = data[6]

    def is_start_stop(self):
        """
        :return: bool
        """
        return self.dimmer_mode == MODE_START_STOP

    def is_dimmer(self):
        """
        :return: bool
        """
        return self.dimmer_mode == MODE_DIMMER

    def is_dimmer_memory(self):
        """
        :return: bool
        """
        return self.dimmer_mode == MODE_MEMORY

    def is_staircase(self):
        """
        :return: bool
        """
        return self.dimmer_mode == MODE_STAIRCASE

    def is_multi(self):
        """
        :return: bool
        """
        return self.dimmer_mode == MODE_MULTI

    def is_slow(self):
        """
        :return: bool
        """
        return self.dimmer_mode == MODE_SLOW

    def is_slow_on(self):
        """
        :return: bool
        """
        return self.dimmer_mode == MODE_SLOW_ON

    def is_slow_off(self):
        """
        :return: bool
        """
        return self.dimmer_mode == MODE_SLOW_OFF

    def cur_dimmer_state(self):
        """
        :return: int
        """
        return self.dimmer_state

    def data_to_binary(self):
        """
        :return: bytes
        """
        return (
            bytes(
                [
                    COMMAND_CODE,
                    self.dimmer_mode,
                    self.dimmer_state,
                    self.led_status,
                ]
            )
            + struct.pack(">L", self.delay_time)[-3:]
        )



"""
:author: Frank van Breugel
"""

from __future__ import annotations

import struct

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xB8
CHANNEL_NORMAL = 0x00
CHANNEL_INHIBITED = 0x01
CHANNEL_FORCED_ON = 0x02
CHANNEL_DISABLED = 0x03

LED_OFF = 0
LED_ON = 1 << 7
LED_SLOW_BLINKING = 1 << 6
LED_FAST_BLINKING = 1 << 5
LED_VERY_FAST_BLINKING = 1 << 4


@register(COMMAND_CODE, ["VMB4DC", "VMBDMI", "VMBDMI-R", "VMB8DC-20", "VMB4LEDPWM-20"])
class DimmerChannelStatusMessage(Message):
    """
    sent by: VMB4DC
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.channel = 1
        self.disable_inhibit_forced = 0
        self.dimmer_state = 0
        self.led_status = 0
        self.delay_time = 0
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 7)
        self.set_attributes(priority, address, rtr)
        self.channel = self.byte_to_channel(data[0])
        self.disable_inhibit_forced = data[1]
        self.dimmer_state = int.from_bytes([data[2]], byteorder="big", signed=False)
        self.led_status = data[3]
        (self.delay_time,) = struct.unpack(">L", bytes([0]) + data[4:])

    def is_normal(self):
        """
        :return: bool
        """
        return self.disable_inhibit_forced == CHANNEL_NORMAL

    def is_inhibited(self):
        """
        :return: bool
        """
        return self.disable_inhibit_forced == CHANNEL_INHIBITED

    def is_forced_on(self):
        """
        :return: bool
        """
        return self.disable_inhibit_forced == CHANNEL_FORCED_ON

    def is_disabled(self):
        """
        :return: bool
        """
        return self.disable_inhibit_forced == CHANNEL_DISABLED

    def cur_dimmer_state(self):
        """
        :return: int
        """
        return self.dimmer_state

    def data_to_binary(self):
        """
        :return: bytes
        """
        return (
            bytes(
                [
                    COMMAND_CODE,
                    self.channels_to_byte([self.channel]),
                    self.disable_inhibit_forced,
                    self.dimmer_state,
                    self.led_status,
                ]
            )
            + struct.pack(">L", self.delay_time)[-3:]
        )



"""
This represents a velbus module
"""

from __future__ import annotations

import importlib.resources
import json
import logging
import os
import pathlib
import struct
import sys
from typing import Awaitable, Callable

from aiofile import async_open

from velbusaio.channels import (
    Blind,
    Button,
    ButtonCounter,
    Channel,
    Dimmer,
    EdgeLit,
    LightSensor,
    Memo,
    Relay,
    SelectedProgram,
    Sensor,
    SensorNumber,
)
from velbusaio.channels import Temperature
from velbusaio.channels import Temperature as TemperatureChannelType
from velbusaio.channels import (
    ThermostatChannel,
)
from velbusaio.command_registry import commandRegistry
from velbusaio.const import (
    CHANNEL_LIGHT_VALUE,
    CHANNEL_MEMO_TEXT,
    CHANNEL_SELECTED_PROGRAM,
    PRIORITY_LOW,
    SCAN_MODULEINFO_TIMEOUT_INITIAL,
)
from velbusaio.helpers import h2, handle_match, keys_exists
from velbusaio.message import Message
from velbusaio.messages.blind_status import BlindStatusMessage, BlindStatusNgMessage
from velbusaio.messages.channel_name_part1 import (
    ChannelNamePart1Message,
    ChannelNamePart1Message2,
    ChannelNamePart1Message3,
)
from velbusaio.messages.channel_name_part2 import (
    ChannelNamePart2Message,
    ChannelNamePart2Message2,
    ChannelNamePart2Message3,
)
from velbusaio.messages.channel_name_part3 import (
    ChannelNamePart3Message,
    ChannelNamePart3Message2,
    ChannelNamePart3Message3,
)
from velbusaio.messages.channel_name_request import (
    COMMAND_CODE as CHANNEL_NAME_REQUEST_COMMAND_CODE,
)
from velbusaio.messages.channel_name_request import (
    ChannelNameRequestMessage,
)
from velbusaio.messages.clear_led import ClearLedMessage
from velbusaio.messages.counter_status import CounterStatusMessage
from velbusaio.messages.counter_status_request import CounterStatusRequestMessage
from velbusaio.messages.counter_value import CounterValueMessage
from velbusaio.messages.dali_device_settings import DaliDeviceSettingMsg
from velbusaio.messages.dali_device_settings import DeviceType as DaliDeviceType
from velbusaio.messages.dali_device_settings import DeviceTypeMsg as DaliDeviceTypeMsg
from velbusaio.messages.dali_device_settings import MemberOfGroupMsg
from velbusaio.messages.dali_device_settings_request import (
    COMMAND_CODE as DALI_DEVICE_SETTINGS_REQUEST_COMMAND_CODE,
)
from velbusaio.messages.dali_device_settings_request import (
    DaliDeviceSettingsRequest,
)
from velbusaio.messages.dali_dim_value_status import DimValueStatus
from velbusaio.messages.dimmer_channel_status import DimmerChannelStatusMessage
from velbusaio.messages.dimmer_status import DimmerStatusMessage
from velbusaio.messages.fast_blinking_led import FastBlinkingLedMessage
from velbusaio.messages.memory_data import MemoryDataMessage
from velbusaio.messages.module_status import (
    ModuleStatusGP4PirMessage,
    ModuleStatusMessage,
    ModuleStatusMessage2,
    ModuleStatusPirMessage,
)
from velbusaio.messages.module_status_request import ModuleStatusRequestMessage
from velbusaio.messages.push_button_status import PushButtonStatusMessage
from velbusaio.messages.raw import MeteoRawMessage, SensorRawMessage
from velbusaio.messages.read_data_from_memory import ReadDataFromMemoryMessage
from velbusaio.messages.relay_status import RelayStatusMessage, RelayStatusMessage2
from velbusaio.messages.sensor_temperature import SensorTemperatureMessage
from velbusaio.messages.set_led import SetLedMessage
from velbusaio.messages.slider_status import SliderStatusMessage
from velbusaio.messages.slow_blinking_led import SlowBlinkingLedMessage
from velbusaio.messages.temp_sensor_status import TempSensorStatusMessage
from velbusaio.messages.update_led_status import UpdateLedStatusMessage


class Module:
    """
    Abstract class for Velbus hardware modules.
    """

    @classmethod
    def factory(
        cls,
        module_address: int,
        module_type: int,
        serial: int | None = None,
        memorymap: int | None = None,
        build_year: int | None = None,
        build_week: int | None = None,
        cache_dir: str | None = None,
    ) -> Module:
        if module_type == 0x45 or module_type == 0x5A:
            return VmbDali(
                module_address,
                module_type,
                serial,
                memorymap,
                build_year,
                build_week,
                cache_dir,
            )

        return Module(
            module_address,
            module_type,
            serial,
            memorymap,
            build_year,
            build_week,
            cache_dir,
        )

    def __init__(
        self,
        module_address: int,
        module_type: int,
        serial: int | None = None,
        memorymap: int | None = None,
        build_year: int | None = None,
        build_week: int | None = None,
        cache_dir: str | None = None,
    ) -> None:
        self._address = module_address
        self._type = int(module_type)
        self._data = {}

        self._name = {}
        self._sub_address = {}
        self.serial = serial
        self.memory_map_version = memorymap
        self.build_year = build_year
        self.build_week = build_week
        self._cache_dir = cache_dir
        self._is_loading = False
        self._channels = {}
        self.loaded = False

    def get_initial_timeout(self) -> int:
        return SCAN_MODULEINFO_TIMEOUT_INITIAL

    async def initialize(self, writer: Callable[[Message], Awaitable[None]]) -> None:
        self._log = logging.getLogger("velbus-module")
        # load the protocol data
        try:
            if sys.version_info >= (3, 13):
                with importlib.resources.path(
                    __name__, f"module_spec/{h2(self._type)}.json"
                ) as fspath:
                    async with async_open(fspath) as protocol_file:
                        self._data = json.loads(await protocol_file.read())
            else:
                async with async_open(
                    str(
                        importlib.resources.files(__name__.split(".")[0]).joinpath(
                            f"module_spec/{h2(self._type)}.json"
                        )
                    )
                ) as protocol_file:
                    self._data = json.loads(await protocol_file.read())
            self._log.debug(f"Module spec {h2(self._type)} loaded")
        except FileNotFoundError:
            self._log.warning(f"No module spec for {h2(self._type)}")
            self._data = {}

        # set some params from the velbus controller
        self._writer = writer
        for chan in self._channels.values():
            chan._writer = writer

    def cleanupSubChannels(self) -> None:
        # TODO: 21/11/2022 DannyDeGaspari: Fix needed
        # Care should be taken for this function, not all subaddresses have their channels on multiples of 8.
        # The last subaddress contain typically the temperature channels, has more then 8 channels
        # and doesn't start on a boundary of 8.
        # E.g. The VMBGP4 has one subaddress, so since the second subaddress is not defined,
        # this function will delete channels 17-24 while 17 and 18 belong to the temperature channels.
        #
        # The solution would be that this functions knows were the temperature channels are located
        # and/or what the max number of subaddresses are for each module.
        # if self._sub_address == {} and self.loaded:
        #   raise Exception("No subaddresses defined")
        for sub in range(1, 4):
            if sub not in self._sub_address:
                for i in range(((sub * 8) + 1), (((sub + 1) * 8) + 1)):
                    if i in self._channels and not isinstance(
                        self._channels[i], TemperatureChannelType
                    ):
                        del self._channels[i]

    async def _cache(self) -> None:
        cfile = pathlib.Path(f"{self._cache_dir}/{self._address}.json")
        async with async_open(cfile, "w") as fl:
            await fl.write(json.dumps(self.to_cache(), indent=4))

    def __getstate__(self) -> dict:
        d = self.__dict__
        self_dict = {k: d[k] for k in d if k != "_writer" and k != "_log"}
        return self_dict

    def __setstate__(self, state: dict) -> None:
        self.__dict__ = state

    def __repr__(self) -> str:
        return f"<{self._name} type:{self._type} address:{self._address} loaded:{self.loaded} loading:{self._is_loading} channels: {self._channels}>"

    def __str__(self) -> str:
        return self.__repr__()

    def to_cache(self) -> dict:
        d = {"name": self._name, "channels": {}}
        for num, chan in self._channels.items():
            d["channels"][num] = chan.to_cache()
        return d

    def get_addresses(self) -> list:
        """
        Get all addresses for this module
        """
        res = []
        res.append(self._address)
        for addr in self._sub_address.values():
            res.append(addr)
        return res

    def get_type(self) -> int:
        """
        Get the module type
        """
        return self._type

    def get_type_name(self) -> str:
        if "Type" in self._data:
            return self._data["Type"]
        return "UNKNOWN"

    def get_serial(self) -> str | None:
        return self.serial

    def get_name(self) -> str:
        return self._name

    def get_sw_version(self) -> str:
        return f"{self.serial}-{self.memory_map_version}.{self.build_year}.{self.build_week}"

    def calc_channel_offset(self, address: int) -> int:
        _channel_offset = 0
        if self._address != address:
            for _sub_addr_key, _sub_addr_val in self._sub_address.items():
                if _sub_addr_val == address:
                    _channel_offset = 8 * _sub_addr_key
                    break
        return _channel_offset

    async def on_message(self, message: Message) -> None:
        """
        Process received message
        """
        self._log.debug(f"RX: {message}")
        _channel_offset = self.calc_channel_offset(message.address)

        if isinstance(
            message,
            (
                ChannelNamePart1Message,
                ChannelNamePart1Message2,
                ChannelNamePart1Message3,
            ),
        ):
            self._process_channel_name_message(1, message)
            await self._cache()
        elif isinstance(
            message,
            (
                ChannelNamePart2Message,
                ChannelNamePart2Message2,
                ChannelNamePart2Message3,
            ),
        ):
            self._process_channel_name_message(2, message)
            await self._cache()
        elif isinstance(
            message,
            (
                ChannelNamePart3Message,
                ChannelNamePart3Message2,
                ChannelNamePart3Message3,
            ),
        ):
            self._process_channel_name_message(3, message)
            await self._cache()
        elif isinstance(message, MemoryDataMessage):
            await self._process_memory_data_message(message)
        elif isinstance(message, (RelayStatusMessage, RelayStatusMessage2)):
            await self._update_channel(
                message.channel,
                {
                    "on": message.is_on(),
                    "inhibit": message.is_inhibited(),
                    "forced_on": message.is_forced_on(),
                    "disabled": message.is_disabled(),
                },
            )
        elif isinstance(message, SensorTemperatureMessage):
            chan = self._translate_channel_name(self._data["TemperatureChannel"])
            await self._channels[chan].maybe_update_temperature(
                message.getCurTemp(), 1 / 64
            )
            await self._update_channel(
                chan,
                {
                    "min": message.getMinTemp(),
                    "max": message.getMaxTemp(),
                },
            )
        elif isinstance(message, TempSensorStatusMessage):
            # update the current temp
            chan = self._translate_channel_name(self._data["TemperatureChannel"])
            if chan in self._channels:
                await self._update_channel(
                    chan,
                    {
                        "target": message.target_temp,
                        "cmode": message.mode_str,
                        "cstatus": message.status_str,
                        "sleep_timer": message.sleep_timer,
                        "cool_mode": message.cool_mode,
                    },
                )
                await self._channels[chan].maybe_update_temperature(
                    message.current_temp, 1 / 2
                )
            # update the thermostat channels
            channel_name_to_msg_prop_map = {
                "Heater": "heater",
                "Boost": "boost",
                "Pump": "pump",
                "Cooler": "cooler",
                "Alarm 1": "alarm1",
                "Alarm 2": "alarm2",
                "Alarm 3": "alarm3",
                "Alarm 4": "alarm4",
            }
            for channel_str in self._data["Channels"]:
                if keys_exists(self._data, "Channels", channel_str, "Type"):
                    if (
                        self._data["Channels"][channel_str]["Type"]
                        == "ThermostatChannel"
                    ):
                        channel = self._translate_channel_name(channel_str)
                        channel_name = self._data["Channels"][channel_str]["Name"]
                        if channel in self._channels:
                            await self._update_channel(
                                channel,
                                {
                                    "closed": getattr(
                                        message,
                                        channel_name_to_msg_prop_map[channel_name],
                                    )
                                },
                            )
        elif isinstance(message, PushButtonStatusMessage):
            _update_buttons = False
            for channel_types in self._data["Channels"]:
                if keys_exists(self._data, "Channels", channel_types, "Type"):
                    if (
                        self._data["Channels"][channel_types]["Type"] == "Button"
                        or self._data["Channels"][channel_types]["Type"] == "Sensor"
                        or self._data["Channels"][channel_types]["Type"]
                        == "ButtonCounter"
                    ):
                        _update_buttons = True
                        break
            if _update_buttons:
                for channel_id in range(1, 9):
                    channel = self._translate_channel_name(channel_id + _channel_offset)
                    if channel_id in message.closed:
                        await self._update_channel(channel, {"closed": True})
                    if channel_id in message.closed_long:
                        await self._update_channel(channel, {"long": True})
                    if channel_id in message.opened:
                        await self._update_channel(
                            channel, {"closed": False, "long": False}
                        )
        elif isinstance(message, (ModuleStatusMessage)):
            for channel_id in range(1, 9):
                channel = self._translate_channel_name(channel_id + _channel_offset)
                if channel_id in message.closed:
                    await self._update_channel(channel, {"closed": True})
                elif channel in self._channels and isinstance(
                    self._channels[channel], (Button, ButtonCounter)
                ):
                    await self._update_channel(channel, {"closed": False})
        elif isinstance(message, (ModuleStatusMessage2)):
            for channel_id in range(1, 9):
                channel = self._translate_channel_name(channel_id + _channel_offset)
                if channel_id in message.closed:
                    await self._update_channel(channel, {"closed": True})
                elif isinstance(self._channels[channel], (Button, ButtonCounter)):
                    await self._update_channel(channel, {"closed": False})
                if channel_id in message.enabled:
                    await self._update_channel(channel, {"enabled": True})
                elif channel in self._channels and isinstance(
                    self._channels[channel], (Button, ButtonCounter)
                ):
                    await self._update_channel(channel, {"enabled": False})
            # self.selected_program_str = message.selected_program_str
            await self._update_channel(
                CHANNEL_SELECTED_PROGRAM,
                {"selected_program_str": message.selected_program_str},
            )
        elif isinstance(message, CounterStatusMessage) and isinstance(
            self._channels[message.channel], ButtonCounter
        ):
            channel = self._translate_channel_name(message.channel)
            await self._update_channel(
                channel,
                {
                    "pulses": message.pulses,
                    "counter": message.counter,
                    "delay": message.delay,
                },
            )
        elif isinstance(message, ModuleStatusPirMessage):
            await self._update_channel(
                CHANNEL_LIGHT_VALUE, {"cur": message.light_value}
            )
            await self._update_channel(1, {"closed": message.dark})
            await self._update_channel(2, {"closed": message.light})
            await self._update_channel(3, {"closed": message.motion1})
            await self._update_channel(4, {"closed": message.light_motion1})
            await self._update_channel(5, {"closed": message.motion2})
            await self._update_channel(6, {"closed": message.light_motion2})
            if 7 in self._channels:
                await self._update_channel(7, {"closed": message.low_temp_alarm})
            if 8 in self._channels:
                await self._update_channel(8, {"closed": message.high_temp_alarm})
            # self.selected_program_str = message.selected_program_str
            await self._update_channel(
                CHANNEL_SELECTED_PROGRAM,
                {"selected_program_str": message.selected_program_str},
            )
        elif isinstance(message, ModuleStatusGP4PirMessage):
            await self._update_channel(
                CHANNEL_LIGHT_VALUE, {"cur": message.light_value}
            )
            for channel_id in range(1, 9):
                channel = self._translate_channel_name(channel_id + _channel_offset)
                await self._update_channel(
                    channel, {"closed": channel_id in message.closed}
                )
                if type(self._channels[channel]) is Button:
                    # only treat 'enabled' if the channel is a Button
                    await self._update_channel(
                        channel, {"enabled": channel_id in message.enabled}
                    )
            # self.selected_program_str = message.selected_program_str
            await self._update_channel(
                CHANNEL_SELECTED_PROGRAM,
                {"selected_program_str": message.selected_program_str},
            )
        elif isinstance(message, UpdateLedStatusMessage):
            for channel_id in range(1, 9):
                channel = self._translate_channel_name(channel_id + _channel_offset)
                if channel_id in message.led_slow_blinking:
                    await self._update_channel(channel, {"led_state": "slow"})
                if channel_id in message.led_fast_blinking:
                    await self._update_channel(channel, {"led_state": "fast"})
                if channel_id in message.led_on:
                    await self._update_channel(channel, {"led_state": "on"})
                if (
                    channel_id not in message.led_slow_blinking
                    and channel_id not in message.led_fast_blinking
                    and channel_id not in message.led_on
                ):
                    await self._update_channel(channel, {"led_state": "off"})
        elif isinstance(message, SetLedMessage):
            for channel_id in range(1, 9):
                channel = self._translate_channel_name(channel_id + _channel_offset)
                if channel_id in message.leds:
                    await self._update_channel(channel, {"led_state": "on"})
        elif isinstance(message, ClearLedMessage):
            for channel_id in range(1, 9):
                channel = self._translate_channel_name(channel_id + _channel_offset)
                if channel_id in message.leds:
                    await self._update_channel(channel, {"led_state": "off"})
        elif isinstance(message, SlowBlinkingLedMessage):
            for channel_id in range(1, 9):
                channel = self._translate_channel_name(channel_id + _channel_offset)
                if channel_id in message.leds:
                    await self._update_channel(channel, {"led_state": "slow"})
        elif isinstance(message, FastBlinkingLedMessage):
            for channel_id in range(1, 9):
                channel = self._translate_channel_name(channel_id + _channel_offset)
                if channel_id in message.leds:
                    await self._update_channel(channel, {"led_state": "fast"})
        elif isinstance(message, (DimmerChannelStatusMessage, DimmerStatusMessage)):
            channel = self._translate_channel_name(message.channel)
            await self._update_channel(channel, {"state": message.cur_dimmer_state()})
        elif isinstance(message, SliderStatusMessage):
            channel = self._translate_channel_name(message.channel)
            await self._update_channel(channel, {"state": message.cur_slider_state()})
        elif isinstance(message, BlindStatusNgMessage):
            channel = self._translate_channel_name(message.channel)
            await self._update_channel(
                channel, {"state": message.status, "position": message.position}
            )
        elif isinstance(message, BlindStatusMessage):
            channel = self._translate_channel_name(message.channel)
            await self._update_channel(channel, {"state": message.status})
        elif isinstance(message, MeteoRawMessage):
            await self._update_channel(11, {"cur": message.rain})
            await self._update_channel(12, {"cur": message.light})
            await self._update_channel(13, {"cur": message.wind})
        elif isinstance(message, SensorRawMessage):
            await self._update_channel(
                message.sensor, {"cur": message.value, "unit": message.unit}
            )
        elif isinstance(message, CounterValueMessage):
            await self._update_channel(
                message.channel, {"power": message.power, "energy": message.energy}
            )

    async def _update_channel(self, channel: int, updates: dict):
        try:
            await self._channels[channel].update(updates)
        except KeyError:
            self._log.info(
                f"channel {channel} does not exist for module @ address {self}"
            )

    def get_channels(self) -> dict:
        """
        List all channels for this module
        """
        return self._channels

    async def load(self, from_cache: bool = False) -> None:
        # start the loading
        self._is_loading = True
        # see if we have a cache
        try:
            cfile = pathlib.Path(f"{self._cache_dir}/{self._address}.json")
            async with async_open(cfile, "r") as fl:
                cache = json.loads(await fl.read())
        except OSError:
            cache = {}
        # load default channels
        await self._load_default_channels()

        # load the data from memory ( the stuff that we need)
        if "name" in cache and cache["name"] != "":
            self._name = cache["name"]
        else:
            await self.__load_memory()
        # load the module status
        # await self._request_module_status()
        # load the channel names
        if "channels" in cache:
            for num, chan in cache["channels"].items():
                self._channels[int(num)]._name = chan["name"]
                if "sub_device" in chan:
                    self._channels[int(num)]._sub_device = chan["sub_device"]
                else:
                    self._channels[int(num)]._sub_device = False
                if "Unit" in chan:
                    self._channels[int(num)]._Unit = chan["Unit"]
                self._channels[int(num)]._is_loaded = True
        else:
            await self._request_channel_name()
        # load the module specific stuff
        self._load()
        # stop the loading
        self._is_loading = False
        await self._request_module_status()

    def _load(self) -> None:
        """
        Method for per module type loading
        """
        pass

    def number_of_channels(self) -> int:
        """
        Retrieve the number of available channels in this module

        :return: int
        """
        if not len(self._channels):
            return 0
        return max(self._channels.keys())

    async def set_memo_text(self, txt: str) -> None:
        if CHANNEL_MEMO_TEXT not in self._channels.keys():
            return
        await self._channels[CHANNEL_MEMO_TEXT].set(txt)

    async def _process_memory_data_message(self, message: MemoryDataMessage) -> None:
        addr = "{high:02X}{low:02X}".format(
            high=message.high_address, low=message.low_address
        )
        if "Memory" not in self._data:
            return
        if "Address" not in self._data["Memory"]:
            return
        mdata = self._data["Memory"]["Address"][addr]
        if "ModuleName" in mdata and isinstance(self._name, dict):
            # if self._name is a dict we are still loading
            # if its a string it was already complete
            char_and_save = mdata["ModuleName"].split(":")
            char = char_and_save[0]
            self._name[int(char)] = chr(message.data)
            if len(char_and_save) > 1 and char_and_save[1] == "Save":
                self._name = "".join(
                    str(x) for x in self._name.values() if x != chr(0xFF)
                )
        elif "Match" in mdata:
            for chan, chan_data in handle_match(mdata["Match"], message.data).items():
                data = chan_data.copy()
                await self._update_channel(chan, data)
        elif "SensorName" in mdata:
            # this is part of the channel names
            # make sure we set the channel to loaded
            # format of the value (in mdata)
            #   channel:char:start/save
            spl = mdata["SensorName"].split(":")
            if len(spl) == 2:
                [chan, pos] = spl
            elif len(spl) == 3:
                [chan, pos, dummy] = spl
            chan = self._translate_channel_name(chan)
            self._channels[chan].set_name_char(pos, message.data)
        else:
            self._log.debug(mdata)

    def _process_channel_name_message(self, part: int, message: Message) -> None:
        channel = self._translate_channel_name(message.channel)
        if channel not in self._channels:
            return
        self._channels[channel].set_name_part(part, message.name)

    def _translate_channel_name(self, channel: str) -> int:
        if keys_exists(
            self._data,
            "ChannelNumbers",
            "Name",
            "Map",
            f"{int(channel):02X}",
        ):
            return int(
                self._data["ChannelNumbers"]["Name"]["Map"][f"{int(channel):02X}"]
            )
        return int(channel)

    async def is_loaded(self) -> bool:
        """
        Check if all name messages have been received
        """
        # if we are loaded, just return
        if self.loaded:
            return True
        if self._is_loading:
            return False
        # the name should be loaded
        if isinstance(self._name, dict):
            return False
        # all channel names should be loaded
        for chan in self._channels.values():
            if not chan.is_loaded():
                return False
        # set that  we finished the module loading
        self.loaded = True
        await self._cache()
        return True

    async def _request_module_status(self) -> None:
        """Request current state of channels."""
        if "Channels" not in self._data:
            # some modules have no channels
            return
        self._log.info(f"Request module status {self._address}")

        mod_stat_req_msg = ModuleStatusRequestMessage(self._address)
        counter_msg = None
        for chan, chan_data in self._data["Channels"].items():
            if int(chan) < 9 and chan_data["Type"] in ("Blind", "Dimmer", "Relay"):
                mod_stat_req_msg.channels.append(int(chan))
            if chan_data["Type"] == "ButtonCounter":
                if counter_msg is None:
                    counter_msg = CounterStatusRequestMessage(self._address)
                counter_msg.channels.append(int(chan))
        await self._writer(mod_stat_req_msg)
        if counter_msg is not None:
            await self._writer(counter_msg)

    async def _request_channel_name(self) -> None:
        # request the module channel names
        if keys_exists(self._data, "AllChannelStatus"):
            msg = ChannelNameRequestMessage(self._address)
            msg.priority = PRIORITY_LOW
            msg.channels = 0xFF
            await self._writer(msg)
        else:
            msg_type = commandRegistry.get_command(
                CHANNEL_NAME_REQUEST_COMMAND_CODE, self.get_type()
            )
            msg = msg_type(self._address)
            msg.priority = PRIORITY_LOW
            msg.channels = list(range(1, (self.number_of_channels() + 1)))
            await self._writer(msg)

    async def __load_memory(self) -> None:
        """
        Request all needed memory addresses
        """
        if "Memory" not in self._data:
            self._name = None
            return

        if self._type == 0x0C:
            self._name = None
            return

        for memory_key, memory_part in self._data["Memory"].items():

            if memory_key == "Address":
                for addr_int in memory_part.keys():
                    addr = struct.unpack(
                        ">BB", struct.pack(">h", int("0x" + addr_int, 0))
                    )
                    msg = ReadDataFromMemoryMessage(self._address)
                    msg.priority = PRIORITY_LOW
                    msg.high_address = addr[0]
                    msg.low_address = addr[1]
                    await self._writer(msg)

    async def _load_default_channels(self) -> None:
        if "Channels" not in self._data:
            return

        for chan, chan_data in self._data["Channels"].items():
            edit = True
            sub = True
            if "Editable" not in chan_data or chan_data["Editable"] != "yes":
                edit = False
            if "Subdevice" not in chan_data or chan_data["Subdevice"] != "yes":
                sub = False
            cls = getattr(sys.modules[__name__], chan_data["Type"])
            self._channels[int(chan)] = cls(
                module=self,
                num=int(chan),
                name=chan_data["Name"],
                nameEditable=edit,
                subDevice=sub,
                writer=self._writer,
                address=self._address,
            )
            if chan_data["Type"] == "Temperature":
                if "Thermostat" in self._data or (
                    "ThermostatAddr" in self._data and self._data["ThermostatAddr"] != 0
                ):
                    await self._update_channel(int(chan), {"thermostat": True})
            if chan_data["Type"] == "Dimmer" and "sliderScale" in self._data:
                self._channels[int(chan)].slider_scale = self._data["sliderScale"]


class VmbDali(Module):
    """
    DALI has a variable number of channels: the number of channels
    depends on the number of DALI devices on the DALI bus
    """

    def __init__(
        self,
        module_address: int,
        module_type: int,
        serial: int | None = None,
        memorymap: int | None = None,
        build_year: int | None = None,
        build_week: int | None = None,
        cache_dir: str | None = None,
    ) -> None:
        super().__init__(
            module_address,
            module_type,
            serial,
            memorymap,
            build_year,
            build_week,
            cache_dir,
        )
        self.group_members: dict[int, set[int]] = {}

    def get_initial_timeout(self) -> int:
        return 100000

    async def _load_default_channels(self) -> None:
        for chan in range(1, 64 + 1):
            self._channels[chan] = Channel(
                module=self,
                num=chan,
                name="placeholder",
                nameEditable=True,
                subDevice=True,
                writer=self._writer,
                address=self._address,
            )
            # Placeholders will keep this module loading
            # Until the DaliDeviceSettings messages either delete or replace these placeholder's
            # with actual channels
        await self._request_dali_channels()

    async def _request_dali_channels(self):
        msg_type = commandRegistry.get_command(
            DALI_DEVICE_SETTINGS_REQUEST_COMMAND_CODE, self.get_type()
        )
        msg: DaliDeviceSettingsRequest = msg_type(self._address)
        msg.priority = PRIORITY_LOW
        msg.channel = 81  # all
        msg.settings = None  # all
        await self._writer(msg)

    async def on_message(self, message: Message) -> None:
        if isinstance(message, DaliDeviceSettingMsg):
            if isinstance(message.data, DaliDeviceTypeMsg):
                if message.data.device_type == DaliDeviceType.NoDevicePresent:
                    if message.channel in self._channels:
                        del self._channels[message.channel]
                elif message.data.device_type == DaliDeviceType.LedModule:
                    if self._channels.get(message.channel).__class__ != Dimmer:
                        # New or changed type, replace channel:
                        self._channels[message.channel] = Dimmer(
                            self,
                            message.channel,
                            None,
                            True,
                            True,
                            self._writer,
                            self._address,
                            slider_scale=254,
                        )
                        await self._request_single_channel_name(message.channel)

            elif isinstance(message.data, MemberOfGroupMsg):
                for group in range(0, 15 + 1):
                    this_group_members = self.group_members.setdefault(group, set())
                    if message.data.member_of_group[group]:
                        this_group_members.add(message.channel)
                    elif message.channel in this_group_members:
                        this_group_members.remove(message.channel)

        elif isinstance(message, PushButtonStatusMessage):
            _channel_offset = self.calc_channel_offset(message.address)
            for channel in message.opened:
                if _channel_offset + channel > 64:  # ignore groups
                    continue
                await self._update_channel((_channel_offset + channel), {"state": 0})
            # ignore message.closed: we don't know at what dimlevel they're started

        elif isinstance(message, DimValueStatus):
            for offset, dim_value in enumerate(message.dim_values):
                channel = message.channel + offset
                if channel <= 64:  # channel
                    await self._update_channel(channel, {"state": dim_value})
                elif channel <= 80:  # group
                    group_num = channel - 65
                    for chan in self.group_members.get(group_num, []):
                        await self._update_channel(chan, {"state": dim_value})
                else:  # broadcast
                    for chan in self._channels.values():
                        await chan.update({"state": dim_value})

        elif isinstance(
            message,
            (
                SetLedMessage,
                ClearLedMessage,
                FastBlinkingLedMessage,
                SlowBlinkingLedMessage,
            ),
        ):
            pass

        else:
            return await super().on_message(message)

        await self._cache()

    async def _request_channel_name(self) -> None:
        # Channel names are requested after channel scan
        # don't do them here (at initialization time)
        pass

    async def _request_single_channel_name(self, channel_num: int) -> None:
        msg_type = commandRegistry.get_command(
            CHANNEL_NAME_REQUEST_COMMAND_CODE, self.get_type()
        )
        msg = msg_type(self._address)
        msg.priority = PRIORITY_LOW
        msg.channels = channel_num
        await self._writer(msg)



"""
:author: Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xD4


@register(COMMAND_CODE, ["VMBEL1", "VMBEL2", "VMBEL4", "VMBELO"])
class EdgeSetCustomColor(Message):
    """
    send by:
    received by: VMBEL1, VMBEL2, VMBEL4, VMBELO
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.set_defaults(address)
        self.pallet = 31
        self.rgb = False
        self.saturation = 0
        self.red = 0
        self.green = 0
        self.blue = 0

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_no_rtr(rtr)
        self.set_attributes(priority, address, rtr)
        self.pallet = data[0]
        self.rgb = bool(data[1] & 0x80)
        self.saturation = data[1] & 0x7F
        self.red = data[2]
        self.green = data[3]
        self.blue = data[4]

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes(
            [
                COMMAND_CODE,
                self.pallet,
                ((self.rgb << 7) + self.saturation),
                self.red,
                self.green,
                self.blue,
            ]
        )



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xF6


@register(COMMAND_CODE)
class SetLedMessage(Message):
    """
    send by: VMB4RYLD
    received by: VMB6IN
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.leds = []
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 1)
        self.set_attributes(priority, address, rtr)
        self.leds = self.byte_to_channels(data[0])

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes([COMMAND_CODE, self.channels_to_byte(self.leds)])



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xED

PROGRAM_SELECTION = {0: "none", 1: "summer", 2: "winter", 3: "holiday"}


@register(COMMAND_CODE)
class ModuleStatusMessage(Message):
    """
    send by: VMB6IN
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.closed = []
        self.led_on = []
        self.led_slow_blinking = []
        self.led_fast_blinking = []
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 4)
        self.set_attributes(priority, address, rtr)
        self.closed = self.byte_to_channels(data[0])
        self.led_on = self.byte_to_channels(data[1])
        self.led_slow_blinking = self.byte_to_channels(data[2])
        self.led_fast_blinking = self.byte_to_channels(data[3])

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes(
            [
                COMMAND_CODE,
                self.channels_to_byte(self.closed),
                self.channels_to_byte(self.led_on),
                self.channels_to_byte(self.led_slow_blinking),
                self.channels_to_byte(self.led_fast_blinking),
            ]
        )


@register(
    COMMAND_CODE,
    [
        "VMB8PBU",
        "VMB6PBN",
        "VMB2PBN",
        "VMB6PBB",
        "VMBGP1",
        "VMBEL1",
        "VMBGP1-2",
        "VMBGP2",
        "VMBEL2",
        "VMBGP2-2",
        "VMBGP4",
        "VMBEL4",
        "VMBGP4-2",
        "VMBGPO",
        "VMBGPOD",
        "VMBGPOD-2",
        "VMBELO",
        "VMB7IN",
        "VMB6PB-20",
        "VMBEL1-20",
        "VMBEL2-20",
        "VMBEL4-20",
        "VMBELO-20",
        "VMBGP1-20",
        "VMBGP2-20",
        "VMBGP4-20",
        "VMBGPO-20",
        "VMBEL4PIR-20",
        "VMBGP4PIR-20",
        "VMBPSUMNGR-20",
    ],
)
class ModuleStatusMessage2(Message):
    def __init__(self, address=None):
        Message.__init__(self)
        self.closed = []
        self.enabled = []
        self.normal = []
        self.locked = []
        self.programenabled = []
        self.selected_program = 0
        self.selected_program_str = PROGRAM_SELECTION[self.selected_program]

    def populate(self, priority, address, rtr, data):
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 6)
        self.set_attributes(priority, address, rtr)
        self.closed = self.byte_to_channels(data[0])
        self.enabled = self.byte_to_channels(data[1])
        self.normal = self.byte_to_channels(data[2])
        self.locked = self.byte_to_channels(data[3])
        self.programenabled = self.byte_to_channels(data[4])
        self.selected_program = data[5] & 0x03
        self.selected_program_str = PROGRAM_SELECTION[self.selected_program]

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes(
            [
                COMMAND_CODE,
                self.channels_to_byte(self.closed),
                self.channels_to_byte(self.enabled),
                self.channels_to_byte(self.normal),
                self.channels_to_byte(self.locked),
            ]
        )


@register(COMMAND_CODE, ["VMBPIRO", "VMBPIRM", "VMBPIRC", "VMBELPIR"])
class ModuleStatusPirMessage(Message):
    def __init__(self, address=None):
        Message.__init__(self)
        # in data[0]
        self.dark: bool = False  # bit 1
        self.light: bool = False  # bit 2
        self.motion1: bool = False  # bit 3
        self.light_motion1: bool = False  # bit 4
        self.motion2: bool = False  # bit 5
        self.light_motion2: bool = False  # bit 6
        self.low_temp_alarm: bool = False  # bit 7
        self.high_temp_alarm: bool = False  # bit 8
        # in data[1] and data[2]
        self.light_value: int = 0
        # in data[5]
        self.selected_program = 0
        self.selected_program_str = PROGRAM_SELECTION[self.selected_program]

    def populate(self, priority, address, rtr, data):
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 7)
        self.set_attributes(priority, address, rtr)
        self.dark = bool(data[0] & (1 << 0))
        self.light = bool(data[0] & (1 << 1))
        self.motion1 = bool(data[0] & (1 << 2))
        self.light_motion1 = bool(data[0] & (1 << 3))
        self.motion2 = bool(data[0] & (1 << 4))
        self.light_motion2 = bool(data[0] & (1 << 5))
        self.low_temp_alarm = bool(data[0] & (1 << 6))
        self.high_temp_alarm = bool(data[0] & (1 << 7))
        self.light_value = (data[1] << 8) + data[2]
        self.selected_program = data[5] & 0x03
        self.selected_program_str = PROGRAM_SELECTION[self.selected_program]

    def data_to_binary(self):
        """
        :return: bytes
        """
        raise NotImplementedError


@register(COMMAND_CODE, ["VMBGP4PIR", "VMBGP4PIR-2"])
class ModuleStatusGP4PirMessage(Message):
    def __init__(self, address=None):
        Message.__init__(self)
        # in data[0]
        self.closed = []
        self.enabled = []  # only 4 bits
        # self.normal = []
        self.locked = []
        self.programenabled = []
        self.selected_program = 0
        self.selected_program_str = PROGRAM_SELECTION[self.selected_program]

        # in data[1] and data[2]
        self.light_value: int = 0
        # in data[5]
        self.selected_program = 0
        self.selected_program_str = PROGRAM_SELECTION[self.selected_program]
        # in data[6]
        self.light_value_send_interval = 0

    def populate(self, priority, address, rtr, data):
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 7)
        self.set_attributes(priority, address, rtr)
        self.closed = self.byte_to_channels(data[0])
        self.enabled = self.byte_to_channels(data[1])
        self.locked = self.byte_to_channels(data[3])
        self.light_value = ((data[1] & 0x30) << 4) + data[2]
        self.programenabled = self.byte_to_channels(data[4])
        self.selected_program = data[5] & 0x03
        self.selected_program_str = PROGRAM_SELECTION[self.selected_program]
        self.light_value_send_interval = data[6]

    def data_to_binary(self):
        """
        :return: bytes
        """
        raise NotImplementedError



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

from velbusaio.message import Message


class ModuleTypeRequestMessage(Message):
    """
    send by:
    received by: VMB6IN, VMB4RYLD
    """

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_rtr(rtr)
        self.needs_no_data(data)
        self.set_attributes(priority, address, rtr)

    def set_defaults(self, address):
        if address is not None:
            self.set_address(address)
        self.set_low_priority()
        self.set_rtr()

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes([])



"""
Helper functions
"""

from __future__ import annotations

import os
import re
from typing import Any

from velbusaio.const import CACHEDIR


def keys_exists(element: dict[str, Any], *keys) -> dict:
    """
    Check if *keys (nested) exists in `element` (dict).
    """
    if not isinstance(element, dict):
        raise AttributeError("keys_exists() expects dict as first argument.")
    if len(keys) == 0:
        raise AttributeError("keys_exists() expects at least two arguments, one given.")

    _element = element
    for key in keys:
        try:
            _element = _element[key]
        except KeyError:
            return False
    return _element


def h2(inp: int) -> str:
    """
    Format as hex uppercase
    """
    return format(inp, "02x").upper()


def handle_match(match_dict: dict[str, dict[str, dict[str, str]]], data: int) -> dict:
    """
    Handle memory match from the module data
    """
    match_result = {}
    binary_data = f"{int(data):08b}"
    for num, match_data in match_dict.items():
        tmp = {}
        for match, res in match_data.items():
            if re.fullmatch(match[1:], binary_data):
                res2 = res.copy()
                res2["Data"] = int(data)
                tmp.update(res2)
        match_result[num] = tmp
    result = {}
    for res in match_result.values():
        if "Channel" in res:
            result[int(res["Channel"])] = {}
            if "SubName" in res and "Value" in res and res["Value"] != "PulsePerUnits":
                result[int(res["Channel"])] = {res["SubName"]: res["Value"]}
            else:
                # Very specifick for vmb7in
                # a = bit 0 to 5 = 0 to 63
                # b = a * 100
                multi = (data & 0x3F) * 100
                # c = bit 6 + 7
                #   00 = x1
                #   01 = x2,5
                #   10 = x0.05
                #   11 = x0.01
                # d = b * c
                if data >> 5 == 3:
                    val = multi * 0.01
                elif data >> 5 == 2:
                    val = multi * 0.05
                elif data >> 5 == 1:
                    val = multi * 2.5
                else:
                    val = multi
                result[int(res["Channel"])] = {res["Value"]: val}
    return result


def get_cache_dir() -> str:
    """Put together the default configuration directory based on the OS."""
    data_dir = os.getenv("APPDATA") if os.name == "nt" else os.path.expanduser("~")
    return os.path.join(data_dir, CACHEDIR)



"""
:author: Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xD8


@register(COMMAND_CODE)
class SetRealtimeClock(Message):
    """
    received by all modules
    """

    def __init__(self, address=0x00, wday=None, hour=None, min=None) -> None:
        Message.__init__(self)
        self._wday = wday
        self._hour = hour
        self._min = min
        self.set_defaults(address)

    def set_defaults(self, address) -> None:
        if address is not None:
            self.set_address(address)
        self.set_low_priority()
        self.set_no_rtr()

    def populate(self, priority, address, rtr, data) -> None:
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 3)
        self.set_attributes(priority, address, rtr)
        self._wday = data[0]
        self._hour = data[1]
        self._min = data[2]

    def data_to_binary(self) -> bytes:
        """
        :return: bytes
        """
        return bytes([COMMAND_CODE, self._wday, self._hour, self._min])



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

import struct

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0x03


@register(COMMAND_CODE)
class StartRelayTimerMessage(Message):
    """
    send by:
    received by: VMB4RYLD
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.relay_channels = []
        self.delay_time = 0
        self.set_defaults(address)

    def set_defaults(self, address):
        if address is not None:
            self.set_address(address)
        self.set_high_priority()
        self.set_no_rtr()

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_high_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 4)
        self.set_attributes(priority, address, rtr)
        self.relay_channels = self.byte_to_channels(data)
        (self.delay_time,) = struct.unpack(">L", bytes([0]) + data[1:])

    def data_to_binary(self):
        """
        :return: bytes
        """
        return (
            bytes([COMMAND_CODE, self.channels_to_byte(self.relay_channels)])
            + struct.pack(">L", self.delay_time)[-3:]
        )



"""
The velbus abstract message class
"""

from __future__ import annotations

import json

from velbusaio.const import PRIORITY_FIRMWARE, PRIORITY_HIGH, PRIORITY_LOW


class ParserError(Exception):
    """
    Error when invalid message is received
    """


class Message:
    """
    Base Velbus message
    """

    def __init__(self, address: int = 0) -> None:
        self.priority = PRIORITY_LOW
        self.address: int = 0
        self.rtr: bool = False
        self.data = bytearray()
        self.set_defaults(address)

    def set_attributes(self, priority: int, address: int, rtr: bool) -> None:
        self.priority = priority
        self.address = address
        self.rtr = rtr

    def populate(self, priority: int, address: int, rtr: bool, data: int) -> None:
        raise NotImplementedError

    def set_defaults(self, address: int | None) -> None:
        """
        Set defaults

        If a message has different than low priority or NO_RTR set,
        then this method needs override in subclass
        """
        if address is not None:
            self.set_address(address)
        self.set_low_priority()
        self.set_no_rtr()

    def set_address(self, address: int) -> None:
        self.address = address

    def data_to_binary(self) -> bytes:
        raise NotImplementedError()

    def to_json_basic(self) -> dict:
        """
        Create JSON structure with generic attributes
        """
        me = {}
        me["name"] = str(self.__class__.__name__)
        me.update(self.__dict__.copy())
        for key in me.copy():
            if key == "name":
                continue
            if callable(getattr(self, key)) or key.startswith("__"):
                del me[key]
            if isinstance(me[key], (bytes, bytearray)):
                me[key] = str(me[key])
        return me

    def to_json(self) -> str:
        """
        Dump object structure to JSON

        This method should be overridden in subclasses to include more than just generic attributes
        """
        return json.dumps(self.to_json_basic())

    def __str__(self) -> str:
        return self.to_json()

    def byte_to_channels(self, byte: int) -> list[int]:
        # pylint: disable-msg=R0201
        result = []
        for offset in range(0, 8):
            if byte & (1 << offset):
                result.append(offset + 1)
        return result

    def channels_to_byte(self, channels: list[int]) -> int:
        # pylint: disable-msg=R0201
        result = 0
        for offset in range(0, 8):
            if offset + 1 in channels:
                result = result + (1 << offset)
        return result

    def byte_to_channel(self, byte: int) -> int:
        channels = self.byte_to_channels(byte)
        self.needs_one_channel(channels)
        return channels[0]

    def parser_error(self, message: str) -> None:
        raise ParserError(self.__class__.__name__ + " " + message)

    def needs_rtr(self, rtr: bool) -> None:
        if not rtr:
            self.parser_error("needs rtr set")

    def set_rtr(self) -> None:
        self.rtr = True

    def needs_no_rtr(self, rtr: bool) -> None:
        if rtr:
            self.parser_error("does not need rtr set")

    def set_no_rtr(self) -> None:
        self.rtr = False

    def needs_low_priority(self, priority: int) -> None:
        if priority != PRIORITY_LOW:
            self.parser_error("needs low priority set")

    def set_low_priority(self) -> None:
        self.priority = PRIORITY_LOW

    def needs_high_priority(self, priority: int) -> None:
        if priority != PRIORITY_HIGH:
            self.parser_error("needs high priority set")

    def set_high_priority(self) -> None:
        self.priority = PRIORITY_HIGH

    def needs_firmware_priority(self, priority: int) -> None:
        if priority != PRIORITY_FIRMWARE:
            self.parser_error("needs firmware priority set")

    def set_firmware_priority(self) -> None:
        self.priority = PRIORITY_FIRMWARE

    def needs_no_data(self, data: bytes) -> None:
        length = len(data)
        if length != 0:
            self.parser_error("has data included")

    def needs_data(self, data: bytes, length: int) -> None:
        if len(data) < length:
            self.parser_error(
                "needs " + str(length) + " bytes of data have " + str(len(data))
            )

    def needs_fixed_byte(self, byte: int, value: int) -> None:
        if byte != value:
            self.parser_error("expects " + chr(value) + " in byte " + chr(byte))

    def needs_one_channel(self, channels: list[int]) -> None:
        if (
            len(channels) != 1
            or not isinstance(channels[0], int)
            or not channels[0] > 0
            or not channels[0] <= 8
        ):
            self.parser_error("needs exactly one bit set in channel byte")



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0x00


@register(COMMAND_CODE)
class PushButtonStatusMessage(Message):
    """
    send by: VMB6IN, VMB4RYLD
    received by: VMB4RYLD
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.closed = []
        self.opened = []
        self.closed_long = []
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_high_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 3)
        self.set_attributes(priority, address, rtr)
        self.closed = self.byte_to_channels(data[0])
        self.opened = self.byte_to_channels(data[1])
        self.closed_long = self.byte_to_channels(data[2])

    def set_defaults(self, address):
        if address is not None:
            self.set_address(address)
        self.set_high_priority()
        self.set_no_rtr()

    def get_channels(self):
        """
        :return: list
        """
        return self.closed + self.opened

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes(
            [
                COMMAND_CODE,
                self.channels_to_byte(self.closed),
                self.channels_to_byte(self.opened),
                self.channels_to_byte(self.closed_long),
            ]
        )



"""
:author: Niels Laukens
"""

from __future__ import annotations

import enum

from velbusaio.command_registry import register
from velbusaio.message import Message
from velbusaio.messages.dali_device_settings import DaliDeviceSetting

COMMAND_CODE = 0xE7


class DataSource(enum.Enum):
    FromMemory = 0
    FromDaliDevice = 1


@register(COMMAND_CODE, ["VMBDALI", "VMBDALI-20"])
class DaliDeviceSettingsRequest(Message):
    """
    send by:
    received by: VMBDALI

    Note: requesting a single setting for all (81) channels does not work (no response)
    """

    def __init__(self, address: int | None = None):
        super().__init__()
        self.channel: int = None
        self.data_source: DataSource = None
        self.settings: DaliDeviceSetting = None
        self.set_defaults(address)

    def populate(self, priority, address: int, rtr: int, data: bytes) -> None:
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 2)
        self.set_attributes(priority, address, rtr)
        self.channel = data[0]
        self.data_source = DataSource(data[1])
        if len(data) >= 3:
            self.settings = data[2]
        else:
            self.settings = None  # all

    def data_to_binary(self) -> bytes:
        data = bytearray([COMMAND_CODE, self.channel, DataSource.FromMemory.value])
        if self.settings is not None:
            data.append(self.settings.value)
        return bytes(data)



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

from velbusaio.messages.blind_status import BlindStatusMessage, BlindStatusNgMessage
from velbusaio.messages.bus_active import BusActiveMessage
from velbusaio.messages.bus_error_counter_status import BusErrorCounterStatusMessage
from velbusaio.messages.bus_error_counter_status_request import (
    BusErrorStatusRequestMessage,
)
from velbusaio.messages.bus_off import BusOffMessage
from velbusaio.messages.channel_name_part1 import (
    ChannelNamePart1Message,
    ChannelNamePart1Message2,
    ChannelNamePart1Message3,
)
from velbusaio.messages.channel_name_part2 import (
    ChannelNamePart2Message,
    ChannelNamePart2Message2,
    ChannelNamePart2Message3,
)
from velbusaio.messages.channel_name_part3 import (
    ChannelNamePart3Message,
    ChannelNamePart3Message2,
    ChannelNamePart3Message3,
)
from velbusaio.messages.channel_name_request import (
    ChannelNameRequestMessage,
    ChannelNameRequestMessage2,
)
from velbusaio.messages.clear_led import ClearLedMessage
from velbusaio.messages.counter_status import CounterStatusMessage
from velbusaio.messages.counter_status_request import CounterStatusRequestMessage
from velbusaio.messages.counter_value import CounterValueMessage
from velbusaio.messages.cover_down import CoverDownMessage, CoverDownMessage2
from velbusaio.messages.cover_off import CoverOffMessage, CoverOffMessage2
from velbusaio.messages.cover_position import CoverPosMessage
from velbusaio.messages.cover_up import CoverUpMessage, CoverUpMessage2
from velbusaio.messages.dali_device_settings import DaliDeviceSettingMsg
from velbusaio.messages.dimmer_channel_status import DimmerChannelStatusMessage
from velbusaio.messages.dimmer_status import DimmerStatusMessage
from velbusaio.messages.fast_blinking_led import FastBlinkingLedMessage
from velbusaio.messages.forced_off import ForcedOff
from velbusaio.messages.forced_on import ForcedOn
from velbusaio.messages.interface_status_request import InterfaceStatusRequestMessage
from velbusaio.messages.ir_receiver_status import IRReceiverStatusMessage
from velbusaio.messages.kwh_status import KwhStatusMessage
from velbusaio.messages.light_value_request import LightValueRequest
from velbusaio.messages.memo_text import MemoTextMessage
from velbusaio.messages.memory_data import MemoryDataMessage
from velbusaio.messages.memory_data_block import MemoryDataBlockMessage
from velbusaio.messages.memory_dump_request import MemoryDumpRequestMessage
from velbusaio.messages.module_status import ModuleStatusMessage, ModuleStatusMessage2
from velbusaio.messages.module_status_request import ModuleStatusRequestMessage
from velbusaio.messages.module_subtype import ModuleSubTypeMessage
from velbusaio.messages.module_type import ModuleType2Message, ModuleTypeMessage
from velbusaio.messages.module_type_request import ModuleTypeRequestMessage
from velbusaio.messages.psu_load import PsuLoadMessage
from velbusaio.messages.psu_values import PsuValuesMessage
from velbusaio.messages.push_button_status import PushButtonStatusMessage
from velbusaio.messages.raw import MeteoRawMessage, SensorRawMessage
from velbusaio.messages.read_data_block_from_memory import (
    ReadDataBlockFromMemoryMessage,
)
from velbusaio.messages.read_data_from_memory import ReadDataFromMemoryMessage
from velbusaio.messages.realtime_clock_status_request import RealtimeClockStatusRequest
from velbusaio.messages.receive_buffer_full import ReceiveBufferFullMessage
from velbusaio.messages.receive_ready import ReceiveReadyMessage
from velbusaio.messages.relay_status import RelayStatusMessage
from velbusaio.messages.restore_dimmer import RestoreDimmerMessage
from velbusaio.messages.select_program import SelectProgramMessage
from velbusaio.messages.sensor_temp_request import SensorTempRequest
from velbusaio.messages.sensor_temperature import SensorTemperatureMessage
from velbusaio.messages.set_date import SetDate
from velbusaio.messages.set_daylight_saving import SetDaylightSaving
from velbusaio.messages.set_dimmer import SetDimmerMessage
from velbusaio.messages.set_led import SetLedMessage
from velbusaio.messages.set_realtime_clock import SetRealtimeClock
from velbusaio.messages.set_temperature import SetTemperatureMessage
from velbusaio.messages.slider_status import SliderStatusMessage
from velbusaio.messages.slow_blinking_led import SlowBlinkingLedMessage
from velbusaio.messages.start_relay_blinking_timer import StartRelayBlinkingTimerMessage
from velbusaio.messages.start_relay_timer import StartRelayTimerMessage
from velbusaio.messages.switch_relay_off import SwitchRelayOffMessage
from velbusaio.messages.switch_relay_on import SwitchRelayOnMessage
from velbusaio.messages.switch_to_comfort import SwitchToComfortMessage
from velbusaio.messages.switch_to_day import SwitchToDayMessage
from velbusaio.messages.switch_to_night import SwitchToNightMessage
from velbusaio.messages.switch_to_safe import SwitchToSafeMessage
from velbusaio.messages.temp_sensor_settings_part1 import TempSensorSettingsPart1
from velbusaio.messages.temp_sensor_settings_part2 import TempSensorSettingsPart2
from velbusaio.messages.temp_sensor_settings_part3 import TempSensorSettingsPart3
from velbusaio.messages.temp_sensor_settings_part4 import TempSensorSettingsPart4
from velbusaio.messages.temp_sensor_settings_request import TempSensorSettingsRequest
from velbusaio.messages.temp_sensor_status import TempSensorStatusMessage
from velbusaio.messages.temp_set_cooling import TempSetCoolingMessage
from velbusaio.messages.temp_set_heating import TempSetHeatingMessage
from velbusaio.messages.update_led_status import UpdateLedStatusMessage
from velbusaio.messages.very_fast_blinking_led import VeryFastBlinkingLedMessage
from velbusaio.messages.write_data_to_memory import WriteDataToMemoryMessage
from velbusaio.messages.write_memory_block import WriteMemoryBlockMessage
from velbusaio.messages.write_module_address_and_serial_number import (
    WriteModuleAddressAndSerialNumberMessage,
)

# pylint: disable-msg=C0301



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

import struct

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0x6A


@register(COMMAND_CODE)
class WriteModuleAddressAndSerialNumberMessage(Message):
    """
    send by:
    received by: VMB4RYLD
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.module_type = 0x00
        self.current_serial = 0
        self.module_address = 0x00
        self.new_serial = 0
        self.set_defaults(address)

    def set_defaults(self, address):
        if address is not None:
            self.set_address(address)
        self.set_firmware_priority()
        self.set_no_rtr()

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_firmware_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 6)
        self.set_attributes(priority, address, rtr)
        self.module_type = data[0]
        prefix = bytes([0, 0])
        (self.current_serial,) = struct.unpack(">L", prefix + data[1] + data[2])
        self.module_address = data[3]
        (self.new_serial,) = struct.unpack(">L", prefix + data[4] + data[5])

    def data_to_binary(self):
        """
        :return: bytes
        """
        return (
            chr(COMMAND_CODE)
            + chr(self.module_type)
            + struct.pack(">L", self.current_serial)[2:]
            + chr(self.module_address)
            + struct.pack(">L", self.new_serial)[2:]
        )



from bs4 import BeautifulSoup


class vlpFile:

    def __init__(self, file_path) -> None:
        self._file_path = file_path
        self._soup = None
        self._modules = {}
        self._read()
        self._parse()
        print(self._modules)

    def _read(self) -> None:
        with open(self._file_path) as file:
            xml_content = file.read()
        self._soup = BeautifulSoup(xml_content, "xml")

    def _parse(self) -> None:
        for module in self._soup.find_all("Module"):
            self._parse_module(module)

    def _parse_module(self, module) -> None:
        mod_info = {}
        mod_info["name"] = module.find("Caption").get_text()
        mod_info["addresses"] = module["address"]
        mod_info["build"] = module["build"]
        mod_info["serial"] = module["serial"]
        mod_info["build"] = module["build"]
        mod_info["type"] = module["type"]
        memory = module.find("Memory")
        mod_info["memory"] = memory.get_text()
        self._modules[mod_info["addresses"]] = mod_info



from __future__ import annotations

import asyncio
import binascii
import logging
import typing as t
from asyncio import transports

import backoff

from velbusaio.const import MAXIMUM_MESSAGE_SIZE, MINIMUM_MESSAGE_SIZE, SLEEP_TIME
from velbusaio.raw_message import RawMessage
from velbusaio.raw_message import create as create_message_info


def _on_write_backoff(details):
    logging.debug(
        f"Transport is not open, waiting {details.wait} seconds after {details.tries}"
    )


class VelbusProtocol(asyncio.BufferedProtocol):
    """Handles the Velbus protocol

    This class is expected to be wrapped inside a VelbusConnection class object which will maintain the socket
    and handle auto-reconnects"""

    def __init__(
        self,
        message_received_callback: t.Callable[[RawMessage], t.Awaitable[None]],
        connection_lost_callback=None,
    ) -> None:
        super().__init__()
        self._log = logging.getLogger("velbus-protocol")
        self._message_received_callback = message_received_callback
        self._connection_lost_callback = connection_lost_callback

        # everything for reading from Velbus
        self._buffer = bytearray(MAXIMUM_MESSAGE_SIZE)
        self._buffer_view = memoryview(self._buffer)
        self._buffer_pos = 0

        self._serial_buf = b""
        self.transport = None

        # everything for writing to Velbus
        self._send_queue = asyncio.Queue()
        self._write_transport_lock = asyncio.Lock()
        self._writer_task = None
        self._restart_writer = False
        self.restart_writing()

        self._closing = False

    def connection_made(self, transport: transports.BaseTransport) -> None:
        self.transport = transport
        self._log.info("Connection established to Velbus")

        self._restart_writer = True
        self.restart_writing()

    async def pause_writing(self) -> None:
        """Pause writing."""
        self._restart_writer = False
        if self._writer_task:
            self._send_queue.put_nowait(None)
        await asyncio.sleep(0.1)

    def restart_writing(self) -> None:
        """Resume writing."""
        if self._restart_writer and not self._write_transport_lock.locked():
            self._writer_task = asyncio.ensure_future(
                self._get_message_from_send_queue()
            )
            self._writer_task.add_done_callback(lambda _future: self.restart_writing())

    def close(self) -> None:
        self._closing = True
        self._restart_writer = False
        if self.transport:
            self.transport.close()

    def connection_lost(self, exc: t.Optional[Exception]) -> None:
        self.transport = None

        if self._closing:
            return  # Connection loss was expected, nothing to do here...
        elif exc is None:
            self._log.warning("EOF received from Velbus")
        else:
            self._log.error(f"Velbus connection lost: {exc!r}")

        self.transport = None
        asyncio.ensure_future(self.pause_writing())
        if self._connection_lost_callback:
            self._connection_lost_callback(exc)

    # Everything read-related

    def get_buffer(self, sizehint: int) -> memoryview:
        return self._buffer_view[self._buffer_pos :]

    def data_received(self, data: bytes) -> None:
        """Receive data from the Streaming protocol.
        Called when asyncio.Protocol detects received data from serial port.
        """
        self._serial_buf += data
        self._log.debug(
            "Received {nbytes} bytes from Velbus: {data_hex}".format(
                nbytes=len(data),
                data_hex=binascii.hexlify(self._serial_buf[: len(data)], " "),
            )
        )
        _recheck = True

        while len(self._serial_buf) > MINIMUM_MESSAGE_SIZE and _recheck:
            # try to construct a Velbus message from the buffer

            _remaining_buf = self._serial_buf[MAXIMUM_MESSAGE_SIZE:]
            msg, remaining_data = create_message_info(
                bytearray(self._serial_buf[:MAXIMUM_MESSAGE_SIZE])
            )

            if msg is not None:
                asyncio.ensure_future(self._process_message(msg))
                _recheck = True
            else:
                _recheck = False
            self._serial_buf = bytes(remaining_data) + _remaining_buf

    def buffer_updated(self, nbytes: int) -> None:
        """Receive data from the Buffered Streaming protocol.
        Called when asyncio.BufferedProtocol detects received data from network.
        """
        self._buffer_pos += nbytes
        # self._log.debug(
        #    "Received {nbytes} bytes from Velbus: {data_hex}".format(
        #        nbytes=nbytes,
        #        data_hex=binascii.hexlify(
        #            self._buffer[self._buffer_pos - nbytes : self._buffer_pos], " "
        #        ),
        #    )
        # )

        if self._buffer_pos > MINIMUM_MESSAGE_SIZE:
            # try to construct a Velbus message from the buffer
            msg, remaining_data = create_message_info(self._buffer)

            if msg is not None:
                asyncio.ensure_future(self._process_message(msg))

            self._new_buffer(remaining_data)

    def _new_buffer(self, remaining_data=None) -> None:
        new_buffer = bytearray(MAXIMUM_MESSAGE_SIZE)
        if remaining_data:
            new_buffer[: len(remaining_data)] = remaining_data

        self._buffer = new_buffer
        self._buffer_pos = len(remaining_data) if remaining_data else 0
        self._buffer_view = memoryview(self._buffer)

    async def _process_message(self, msg: RawMessage) -> None:
        # self._log.debug(f"RX: {msg}")
        await self._message_received_callback(msg)

    # Everything write-related

    async def write_auth_key(self, authkey: str) -> None:
        self._log.debug("TX: authentication key")
        if not self.transport.is_closing():
            self.transport.write(authkey.encode("utf-8"))

    async def send_message(self, msg: RawMessage) -> None:
        self._send_queue.put_nowait(msg)

    async def _get_message_from_send_queue(self) -> None:
        self._log.debug("Starting Velbus write message from send queue")
        self._log.debug("Acquiring write lock")
        await self._write_transport_lock.acquire()
        while self._restart_writer:
            # wait for an item from the queue
            msg_info = await self._send_queue.get()
            if msg_info is None:
                self._restart_writer = False
                return
            message_sent = False
            try:
                while not message_sent:
                    message_sent = await self._write_message(msg_info)
                if msg_info.command == 0xEF:
                    # 'channel name request' command provokes in worst case 99 answer packets from VMBGPOD
                    queue_sleep_time = SLEEP_TIME * 33
                else:
                    queue_sleep_time = SLEEP_TIME
                await asyncio.sleep(queue_sleep_time)
            except (asyncio.CancelledError, GeneratorExit) as exc:
                if not self._closing:
                    self._log.error(f"Stopping Velbus writer due to {exc!r}")
                self._restart_writer = False
            except Exception as exc:
                self._log.error(f"Restarting Velbus writer due to {exc!r}")
                self._restart_writer = True
        if self._write_transport_lock.locked():
            self._write_transport_lock.release()
        self._log.debug("Ending Velbus write message from send queue")

    @backoff.on_predicate(
        backoff.expo,
        lambda is_sent: not is_sent,
        max_tries=10,
        on_backoff=_on_write_backoff,
    )
    async def _write_message(self, msg: RawMessage) -> bool:
        self._log.debug(f"TX: {msg}")
        if not self.transport.is_closing():
            self.transport.write(msg.to_bytes())
            return True
        else:
            return False
